# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

SOURCE = """\
define test::is_odd returns bool
    argument int x
    dup
    bne 0, recurse
    load FALSE
    return
recurse:
    sub 1
    call is_even

define test::is_even returns bool
    argument int x
    dup
    bne 0, recurse
    load TRUE
    return
recurse:
    sub 1
    call is_odd
"""

IS_ODD = "test::is_odd(i)i"
IS_EVEN = "test::is_even(i)i"

class TestImplicitExtern(TestCase):
    def test_implicit_extern(self):
        """Test that implicit externals work."""
        tree, output = self.compile(SOURCE)
        for func, input, expect in ((IS_ODD, 0, 0),
                                    (IS_ODD, 1, 1),
                                    (IS_EVEN, 0, 1),
                                    (IS_EVEN, 1, 0),
                                    (IS_ODD, 17, 1),
                                    (IS_EVEN, 23, 0),
                                    (IS_ODD, 38, 0),
                                    (IS_EVEN, 42, 1)):
            self.assertEqual(output.call(func, input), [expect])




# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import ParserError, RedefinedIdentError

SOURCE = """\
define test::load_slot
   argument int an_argument
   extern ptr a_symbol
   extern func () a_function
   extern func () other_provider::function_2
   load a_symbol
   load a_function
   load other_provider::function_2
"""

SOURCE_NAMES_SLOTS = (("an_argument", 3),
                      ("a_symbol", 2),
                      ("a_function", 1),
                      ("test::a_function", 1),
                      ("other_provider::function_2", 0))

SOURCE_WITH_NAME = SOURCE + """\
   load 23
   name 0, a_name
"""

SWN_NAMES_SLOTS = (("an_argument", 4),
                   ("a_symbol", 3),
                   ("a_function", 2),
                   ("test::a_function", 2),
                   ("other_provider::function_2", 1),
                   ("a_name", 0))

class TestLoadSlot(TestCase):
    def test_shortname_only(self):
        """Check that name rejects fullnames."""
        source = SOURCE + "name 2 provider::shortname"
        self.assertRaises(ParserError, self.compile, source)

    def test_basic(self):
        """Check externals' names work."""
        for name, expect_slot in SOURCE_NAMES_SLOTS:
            source = SOURCE + "load " + name
            tree, output = self.compile(source)
            ops = output.ops
            self.assertEqual(len(ops), 4)
            op = ops[-1]
            if expect_slot == 0:
                self.assertEqual(op.name, "dup")
            elif expect_slot == 1:
                self.assertEqual(op.name, "over")
            else:
                self.assertEqual(op.name, "pick")
                self.assertEqual(op.operand, expect_slot)

    def test_name_anonymous_slot(self):
        """Check naming anonymous slots works."""
        for name, expect_slot in SWN_NAMES_SLOTS:
            tree, output = self.compile(SOURCE_WITH_NAME + "load " + name)
            ops = output.ops
            self.assertEqual(len(ops), 5)
            op = ops[-1]
            if expect_slot == 0:
                self.assertEqual(op.name, "dup")
            elif expect_slot == 1:
                self.assertEqual(op.name, "over")
            else:
                self.assertEqual(op.name, "pick")
                self.assertEqual(op.operand, expect_slot)

    def test_name_named_slot(self):
        """Check naming already-named slots works."""
        for name, expect_slot in SOURCE_NAMES_SLOTS:
            names = ["a_name", name]
            for order in 0, 1:
                source = SOURCE + "\n".join(
                    ["name %d, a_name" % expect_slot]
                    + ["load %s" % name for name in names])
                tree, output = self.compile(source)
                ops = output.ops
                self.assertEqual(len(ops), 5)

                # The first load can pick from anywhere
                op = ops[-2]
                if expect_slot == 0:
                    self.assertEqual(op.name, "dup")
                elif expect_slot == 1:
                    self.assertEqual(op.name, "over")
                else:
                    self.assertEqual(op.name, "pick")
                    self.assertEqual(op.operand, expect_slot)

                # The second load should grab the top slot
                self.assertEqual(ops[-1].name, "dup")

                names.reverse()

    def test_shadowing_rename(self):
        """Check that you cannot shadow slots."""
        for name, expect_slot in SWN_NAMES_SLOTS:
            if name.find("::") >= 0:
                continue
            assert expect_slot != 1
            source = SOURCE_WITH_NAME + "name 1, %s" % name
            self.assertRaises(RedefinedIdentError, self.compile, source)

    def test_existing_name(self):
        """Check that giving a slot its existing name is ok."""
        for name, expect_slot in SWN_NAMES_SLOTS:
            if name.find("::") >= 0:
                continue
            source = SOURCE_WITH_NAME + "\n".join(
                ["name %d, %s" % (expect_slot, name),
                 "load %s" % name])
            tree, output = self.compile(source)
            ops = output.ops
            self.assertEqual(len(ops), 5)
            op = ops[-1]
            if expect_slot == 0:
                self.assertEqual(op.name, "dup")
            elif expect_slot == 1:
                self.assertEqual(op.name, "over")
            else:
                self.assertEqual(op.name, "pick")
                self.assertEqual(op.operand, expect_slot)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import strtoint_c
from . import LexerError
import re

TOKEN = re.compile(r"\s+|::|[:,()]|"
                   + r"-?(0x[0-9a-f]+|0[0-7]*|[0-9]+)|"
                   + r"[a-z_][a-z0-9_]*", re.IGNORECASE)

class Token(object):
    def __init__(self, filename, linenumber, text):
        self.filename = filename
        self.linenumber = linenumber
        self.text = text

    @property
    def fileline(self):
        return "%s:%d" % (self.filename, self.linenumber)

    def __str__(self):
        return "%s: %s" % (self.fileline, repr(self.text).lstrip("u"))

class SyntheticToken(Token):
    """A token that the parser inserted.
    """
    def __init__(self, template, text):
        assert isinstance(template, Token)
        Token.__init__(self, template.filename, template.linenumber, text)

class NEWLINE(Token): pass
class COMMA(Token): pass
class OPAREN(Token): pass
class CPAREN(Token): pass
class COLON(Token): pass
class DOUBLE_COLON(Token): pass
class WORD(Token): pass

class NUMBER(Token):
    def __init__(self, *args):
        Token.__init__(self, *args)
        self.value = strtoint_c(self.text, self.__wrap_exception)

    def __wrap_exception(self, msg):
        raise LexerError(self.filename, self.linenumber, msg)

class STRING(Token):
    def __init__(self, *args):
        Token.__init__(self, *args)
        self.value = eval(self.text)

SIMPLE_CLASSES = {
    ",": COMMA,
    "(": OPAREN,
    ")": CPAREN,
    ":": COLON,
    "::": DOUBLE_COLON}

def parse_line_control(line):
    line = line.split()
    return eval(line[2]), int(line[1])

def parse_quoted(line):
    for index, c in zip(range(len(line)), line):
        if index == 0:
            term = c
            skip = False
        elif not skip and c == term:
            return line[:index + 1]
        else:
            if skip and c not in '\\"':
                break # Not supported by emitter.String.quote
            skip = c == '\\'

def generate_tokens(readline):
    filename, linenumber = None, 0
    last_token = None
    while True:
        line = readline()
        if not line:
            break
        line = line.decode("utf-8")
        if line.startswith("#"):
            filename, linenumber = parse_line_control(line)
            continue
        while line:
            tokentext = None
            if line.startswith('"'):
                tokentext = parse_quoted(line)
            else:
                match = TOKEN.match(line)
                if match is not None:
                    tokentext = match.group(0)
            if tokentext is None:
                raise LexerError(filename, linenumber,
                                 "invalid syntax: ‘%s’" % line.rstrip())
            assert len(tokentext) > 0
            assert line.startswith(tokentext)
            line = line[len(tokentext):]
            klass = SIMPLE_CLASSES.get(tokentext, None)
            if klass is None:
                if tokentext[0].isspace():
                    if (not (last_token is None
                             or isinstance(last_token, NEWLINE))
                        and "\n" in tokentext):
                        klass = NEWLINE
                elif tokentext[0] == "-" or tokentext[0].isdigit():
                    klass = NUMBER
                elif tokentext[0] == '"':
                    klass = STRING
                else:
                    klass = WORD
            if klass is not None:
                last_token = klass(filename, linenumber, tokentext)
                yield last_token
            # Hack to allow labels on the same line as operations.
            if klass is COLON:
                last_token = NEWLINE(filename, linenumber, "synthetic")
                yield last_token
        linenumber += 1



# -*- coding: utf-8 -*-
# Copyright (C) 2016 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c import compat
from i8c.compiler import parser
from i8c.compiler import I8CError
from i8c.compiler import LexerError, ParserError
from i8c.compiler import UndefinedIdentError
from i8c.compiler.types import INTTYPE, PTRTYPE, BOOLTYPE

INPUTS = (
    ("0", INTTYPE),
    ("1", INTTYPE),
    ("8", INTTYPE),
    ("9", INTTYPE),

    ("12345", INTTYPE),
    ("01234", INTTYPE),
    ("0x123", INTTYPE),

    ("0o123", None), # Python style octal
    ("0b101", None),
    ("0n123", None),

    ("1a234", None),
    ("01a23", None),
    ("01823", None),
    ("0x1g4", None),

    ("0X123", INTTYPE),
    ("0O123", None),
    ("0B101", None),
    ("0N123", None),
    ("1A234", None),
    ("01A23", None),
    ("01823", None),
    ("0X1G4", None),

    ("NULL", PTRTYPE),
    ("TRUE", BOOLTYPE),
    ("FALSE", BOOLTYPE),

    ("NXLL", None),
    ("TXUE", None),
    ("FXLSE", None),

    ("Null", None),
    ("True", None),
    ("False", None),

    ("null", None),
    ("True", None),
    ("False", None),

    ("NONE", None),
    ("None", None),
    ("none", None))

SOURCE = """\
define test::input_test
    load %s
    drop
"""

class TestNumericConstant(TestCase):
    def test_numeric(self):
        """Check that numeric constants are parsed correctly.
        """
        for input, expect in INPUTS:
            for prefix in "", "-":
                self.__do_test(prefix + input, expect)

    def __do_test(self, input, expect):
        exception = ParserError
        if input.startswith("-"):
            if expect not in (None, INTTYPE):
                expect = None
            if not input[1].isdigit():
                exception = LexerError
        elif expect is None and not input[0].isdigit():
            exception = UndefinedIdentError

        source = SOURCE % input
        if expect is None:
            self.assertRaises(exception, self.compile, source)
            return

        tree, output = self.compile(source)
        node = tree.one_child(parser.Function)
        node = node.one_child(parser.Operations)

        self.assertEqual(len(node.children), 2)
        load, drop = node.children
        self.assertIsInstance(load, parser.LoadOp)
        self.assertIsInstance(drop, parser.SimpleOp)
        self.assertEqual(drop.name, "drop")

        self.assertEqual(len(load.children), 1)
        [node] = load.children
        self.assertIsInstance(node, parser.Constant)
        self.assertEqual(node.type, expect)

    def test_bad_octal(self):
        """Check that compat.strtoint_c catches bad octal.
        """
        for prefix in "", "-":
            for base in range(ord("a"), ord("z") + 1):
                base = chr(base)
                if base == "x":
                    continue
                for base in base.lower(), base.upper():
                    self.assertRaises(I8CError,
                                      compat.strtoint_c,
                                      "%s0%s0" % (prefix, base),
                                      I8CError)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import logger
from . import RedefinedIdentError
from . import UndefinedIdentError
from . import visitors
from . import warn
from .operations import *

debug_print = logger.debug_printer_for(__name__)

class BlockLabel(NonTerminalOp):
    """Placeholder used only during block creation."""

    @property
    def name(self):
        return self.ast.name

class BasicBlock(visitors.Visitable):
    def __init__(self, index):
        self.index = index
        self.label = None
        self.entries = {}
        self.ops = []

    @property
    def fileline(self):
        return self.ops[0].fileline

    @property
    def name(self):
        return "Block #%d (%s)" % (self.index,
                                   (self.label is not None
                                    and '"%s"' % self.label
                                    or self.fileline))

    @property
    def first_op(self):
        return self.ops[0]

    @property
    def last_op(self):
        return self.ops[-1]

    @property
    def is_terminated(self):
        return self.ops and self.last_op.is_block_terminal

    @property
    def is_branch_terminated(self):
        return self.ops and self.last_op.is_branch

    @property
    def branched_exit(self):
        assert self.is_branch_terminated
        assert len(self.exits) == 2
        return self.exits[BranchOp.BRANCHED_EXIT]

    @property
    def nobranch_exit(self):
        assert self.is_branch_terminated
        assert len(self.exits) == 2
        return self.exits[BranchOp.NOBRANCH_EXIT]

    @property
    def is_goto_terminated(self):
        return self.ops and self.last_op.is_goto

    @property
    def goto_exit(self):
        assert self.is_goto_terminated
        assert len(self.exits) == 1
        return self.exits[0]

    @property
    def is_return_terminated(self):
        return self.ops and self.last_op.is_return

    def append(self, op):
        assert not self.is_terminated
        if isinstance(op, BlockLabel):
            assert self.label is None
            self.label = op.name
        self.ops.append(op)

    def set_exits(self, labels):
        try:
            self.exits = [labels[label]
                          for label in self.last_op.exit_labels]
        except KeyError as e:
            raise UndefinedIdentError(self.last_op, "label", e.args[0])

    def __str__(self):
        result = self.name
        if hasattr(self, "exits"):
            result += ", "
            if len(self.exits) == 0:
                result += "no exits"
            elif len(self.exits) == 1:
                result += "next block = #%d" % self.exits[0].index
            else:
                result += "exits = %s" % ", ".join(
                    ("#%d" % block.index for block in self.exits))
        result += ":"
        entry_stack = getattr(self, "entry_stack", None)
        if entry_stack is not None:
            result += "\n%s\n  ---" % entry_stack
        for op in self.ops:
            result += "\n  %s" % op
        return result

    def is_equivalent_to(self, other):
        assert not self is other

        # Quickly reject mismatches
        if other.exits != self.exits:
            return False
        if len(other.ops) != len(self.ops):
            return False
        if other.entry_stack.depth != self.entry_stack.depth:
            return False

        # Check the operations
        for op1, op2 in zip(other.ops, self.ops):
            if not op1.is_equivalent_to(op2):
                return False

        # Check the entry stacks
        for index in range(self.entry_stack.depth):
            bt1 = other.entry_stack[index].basetype
            bt2 = self.entry_stack[index].basetype
            if bt1 != bt2:
                return False

        return True

    def replace_exit(self, old, new):
        self.exits = [block == old and new or block
                      for block in self.exits]

class Label(object):
    is_builtin = False

    def __init__(self, ast, pc):
        self.ast = ast
        self.pc = pc

    @property
    def fileline(self):
        return self.ast.fileline

class SyntheticLabel(Label):
    def __init__(self, pc):
        Label.__init__(self, None, pc)

class BlockCreator(object):
    def visit_toplevel(self, toplevel):
        for node in toplevel.functions:
            node.accept(self)

    def visit_function(self, function):
        self.blocks = {}
        self.labels = {}
        self.__block = None
        self.pc = 0

        function.ops.accept(self)
        del function.ops
        self.ensure_has_blocks(function)
        self.ensure_all_blocks_terminated()

        labels = {}
        for name, label in self.labels.items():
            labels[name] = self.blocks[label.pc]
        del self.labels

        blocks = sorted(self.blocks.items())
        blocks = [block for start_pc, block in blocks]
        del self.blocks

        for block in blocks:
            block.set_exits(labels)
            for exit_block in block.exits:
                exit_block.entries[block] = True
        for block in blocks:
            block.entries = list(block.entries.keys())

        if debug_print.is_enabled:
            for block in blocks:
                debug_print(str(block) + "\n\n")

        function.entry_block = blocks[0]
        function.entry_block.entries.append(None)

        for block in blocks:
            if not block.entries:
                warn("code is unreachable", block)

    def new_synthetic_label(self, target):
        # Create a synthetic label.  Using an integer
        # means it cannot clash with any user-supplied
        # labels as they are all strings.
        label = len(self.labels)
        assert label not in self.labels
        self.labels[label] = SyntheticLabel(target)
        return label

    def ensure_has_blocks(self, function):
        if self.blocks:
            return
        self.add_op(SyntheticReturn(Operation(function)))

    def ensure_all_blocks_terminated(self):
        blocks = sorted(self.blocks.items())
        start_pcs = [start_pc for start_pc, block in blocks]
        blocks = [block for start_pc, block in blocks]
        for block, next_pc in zip(blocks, start_pcs[1:] + [None]):
            op = block.last_op
            if not block.is_terminated:
                if next_pc is None:
                    block.append(SyntheticReturn(op))
                else:
                    label = self.new_synthetic_label(next_pc)
                    block.append(SyntheticGoto(op, label))
            elif block.is_branch_terminated:
                if next_pc is None:
                    label = self.new_synthetic_label(self.pc)
                    self.add_op(SyntheticReturn(op))
                else:
                    label = self.new_synthetic_label(next_pc)
                op.fallthrough = label
            if isinstance(block.first_op, BlockLabel):
                block.ops.pop(0)

    @property
    def current_block(self):
        if self.__block is None:
            assert self.pc not in self.blocks
            self.__block = BasicBlock(len(self.blocks))
            self.blocks[self.pc] = self.__block
        return self.__block

    def drop_current_block(self):
        self.__block = None

    def add_op(self, op):
        block = self.current_block
        block.append(op)
        if block.is_terminated:
            self.drop_current_block()
        self.pc += 1

    def visit_operations(self, ops):
        for child in ops.children:
            child.accept(self)

    def visit_label(self, label):
        prev = self.labels.get(label.name, None)
        if prev is not None:
            raise RedefinedIdentError(label, "label", label.name, prev)
        self.labels[label.name] = Label(label, self.pc)
        self.drop_current_block()
        self.add_op(BlockLabel(label))

    # Operations

    def visit_simpleop(self, op):
        self.add_op(globals()[op.name.title() + "Op"](op))

    def visit_castop(self, op):
        self.add_op(CastOp(op))

    def visit_compareop(self, op):
        self.add_op(CompareOp(op))

    def visit_condbranchop(self, op):
        self.add_op(CompareOp(op))
        self.add_op(BranchOp(op))

    def visit_derefop(self, op):
        self.add_op(DerefOp(op))

    def visit_dupop(self, op):
        self.add_op(PickOp(op, 0))

    def visit_gotoop(self, op):
        self.add_op(GotoOp(op))

    def visit_loadop(self, op):
        if hasattr(op.operand, "type"):
            self.add_op(ConstOp(op))
        else:
            self.add_op(LoadOp(op))

    def visit_nameop(self, op):
        self.add_op(NameOp(op))

    def visit_overop(self, op):
        self.add_op(PickOp(op, 1))

    def visit_pickop(self, op):
        self.add_op(PickOp(op, op.operand.value))

    def visit_returnop(self, op):
        self.add_op(ReturnOp(op))

    def visit_warnop(self, op):
        self.add_op(WarnOp(op))



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import join_bytes
import struct

class Builder(object):
    def __init__(self, mem):
        self.mem = mem

    @property
    def env(self):
        return self.mem.env

    def __enter__(self):
        self.blocks = []
        self.symbols = {}
        return self

    def alloc(self, name=None):
        result = AllocatedBlock(self.env.wordsize)
        self.blocks.append(result)
        if name is not None:
            assert not name in self.symbols
            self.symbols[name] = result
        return result

    def __exit__(self, type, value, traceback):
        if type is None:
            # Allocate block addresses.
            addr = 0x10000000
            for block in self.blocks:
                block.location = addr
                addr += block.length
                addr += 1 # empty space to catch overflows
                addr += (16 - addr % 16) # align for printing
            # Store the data.
            for block in self.blocks:
                block.write_into(self.mem)
            # Install the symbols.
            for name, block in self.symbols.items():
                self.env.register_symbol(name, block.location)
        # Avoid circular references.
        for block in self.blocks:
            del block.fields

class Block(object):
    def __init__(self, offset, fields):
        self.offset = offset
        self.fields = fields

    def __add__(self, offset):
        return OffsetBlock(self, offset)

    def store_i8(self, offset, value):
        self.__store(offset, value, 8, True)

    def store_u8(self, offset, value):
        self.__store(offset, value, 8, False)

    def store_i16(self, offset, value):
        self.__store(offset, value, 16, True)

    def store_u16(self, offset, value):
        self.__store(offset, value, 16, False)

    def store_i32(self, offset, value):
        self.__store(offset, value, 32, True)

    def store_u32(self, offset, value):
        self.__store(offset, value, 32, False)

    def store_i64(self, offset, value):
        self.__store(offset, value, 64, True)

    def store_u64(self, offset, value):
        self.__store(offset, value, 64, False)

    def store_ixx(self, offset, value):
        self.__store(offset, value, self.wordsize, True)

    def store_uxx(self, offset, value):
        self.__store(offset, value, self.wordsize, False)

    def store_ptr(self, offset, value):
        if value in (0, None): # 0 == NULL
            value = 0
        else:
            assert isinstance(value, Block)
        self.store_uxx(offset, value)

    def __store(self, offset, value, nbits, is_signed):
        offset += self.offset
        value = Value(value, nbits, is_signed)
        self.fields[offset] = value
        for i in range(1, value.length):
            self.fields[offset + i] = None

    @property
    def length(self):
        if not self.fields:
            # Really 0, but returning 1 leaves an empty sentinel
            # cell to cause an error in the event this pointer to
            # uninitialized memory is dereferenced.
            return 1
        return max(self.fields.keys()) + 1

    def write_into(self, mem):
        for offset, value in sorted(self.fields.items()):
            if value is not None:
                value.write_into(mem, self.location + offset)

class AllocatedBlock(Block):
    def __init__(self, wordsize):
        Block.__init__(self, 0, {})
        self.wordsize = wordsize
        self.offset = 0

class OffsetBlock(Block):
    def __init__(self, parent, offset):
        Block.__init__(self, parent.offset + offset, parent.fields)
        self.parent = parent

    @property
    def wordsize(self):
        return self.parent.wordsize

    @property
    def location(self):
        root = self
        while isinstance(root, OffsetBlock):
            root = root.parent
        return root.location + self.offset

class Value(object):
    FORMATS = {8: b"b", 16: b"h", 32: b"i", 64: b"q"}

    def __init__(self, value, nbits, is_signed):
        self.value = value
        self.format = self.FORMATS[nbits]
        if not is_signed:
            self.format = self.format.upper()
        self.length = struct.calcsize(self.format)
        assert self.length * 8 == nbits

    def write_into(self, mem, location):
        value = self.value
        if isinstance(value, Block):
            value = value.location
        bytes = struct.pack(mem.env.byteorder + self.format, value)
        mem.write(location, bytes)

class Memory(object):
    def __init__(self, env):
        self.env = env
        self.cells = {}

    def builder(self):
        return Builder(self)

    def getbyte(self, location):
        try:
            return self.cells[location]
        except KeyError:
            # XXX this should be an exception
            print("getbyte(%x):\n%s" % (location, self))
            raise

    def putbyte(self, location, value):
        assert not location in self.cells
        self.cells[location] = value

    def read(self, location, size):
        return join_bytes(self.getbyte(location + offset)
                          for offset in range(size))

    def write(self, location, bytes):
        for byte, offset in zip(bytes, range(len(bytes))):
            self.putbyte(location + offset, byte)

    def __str__(self):
        tmp = {}
        for location, content in sorted(self.cells.items()):
            col = location % 16
            row = location - col
            if not row in tmp:
                tmp[row] = [None] * 16
            tmp[row][col] = content
        lines = []
        locfmt = "  %%0%dx: " % (self.env.wordsize // 4)
        lastloc = None
        for location, cells in sorted(tmp.items()):
            if lastloc not in (None, location - 16):
                lines.append("   ...")
            line = [locfmt % location]
            for cell in cells:
                if cell is None:
                    cell = "--"
                else:
                    cell = "%02x" % cell
                line.append(cell)
            line.insert(9, "")
            lines.append(" ".join(line))
            lastloc = location
        return "\n".join(lines)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

def read_uleb128(data, start):
    return __read_leb128(data, start, False)

def read_sleb128(data, start):
    return __read_leb128(data, start, True)

def __read_leb128(data, start, is_signed):
    result = shift = 0
    while True:
        byte = ord(data[start])
        start += 1
        result |= ((byte & 0x7f) << shift)
        if (byte & 0x80) == 0:
            break
        shift += 7
    if is_signed and (byte & 0x40):
        sign = 0x40 << shift
        result &= ~(0x40 << shift)
        result -= sign
    return start, result



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import strtoint_c
from . import I8XError, HeaderFileError, TestFileError
from . import memory
import copy
import inspect
import os
import platform
import struct
import weakref
try:
    import unittest2 as unittest
except ImportError: # pragma: no cover
    import unittest

class BaseTestCase(unittest.TestCase):
    def run(self, *args, **kwargs):
        self.memory = memory.Memory(self)
        self.addCleanup(delattr, self, "memory")
        self.__symbols = {}
        return unittest.TestCase.run(self, *args, **kwargs)

    def register_symbol(self, name, value):
        """Hook method for registering a symbol."""
        assert not name in self.__symbols
        self.__symbols[name] = value

    def lookup_symbol(self, name):
        """Hook method for resolving a symbol name to an address."""
        return self.__symbols[name]

    def read_memory(self, fmt, addr):
        """Hook method for reading bytes from memory."""
        return self.memory.read(addr, struct.calcsize(fmt))

    def warn_caller(self, msg):
        """Hook method for warning the caller about something."""
        self.fail("unexpected warning: " + msg)

class TestCase(BaseTestCase):
    include_path = []

    @staticmethod
    def __testcase_frame():
        frame = inspect.currentframe()
        if frame is None:
            raise I8XError("not available on "
                           + platform.python_implementation())
        return frame.f_back.f_back

    @classmethod
    def import_builtin_constants(cls):
        frame = cls.__testcase_frame()
        frame.f_globals.update({"NULL": 0, "FALSE": 0, "TRUE": 1})

    @classmethod
    def import_constants_from(cls, filename):
        frame = cls.__testcase_frame()
        path = copy.copy(cls.include_path)
        path.insert(0, os.path.dirname(frame.f_code.co_filename))
        for try_dir in path:
            try_filename = os.path.join(try_dir, filename)
            if os.path.exists(try_filename):
                filename = try_filename
                break
        else:
            raise TestFileError(filename, "not found in: " + repr(path))
        lines = open(filename).readlines()
        for line, linenumber in zip(lines, range(1, len(lines) + 1)):
            bits = line.strip().split()
            try:
                assert len(bits) == 3 and bits[0] == "#define"
                name, value = bits[1:]
                value = strtoint_c(value, I8XError)
                frame.f_globals[name] = value
            except:
                raise HeaderFileError(filename, linenumber)

    def run(self, *args, **kwargs):
        self.addCleanup(self.__restore_env, self.i8ctx.env)
        self.i8ctx.env = self
        return BaseTestCase.run(self, *args, **kwargs)

    def __restore_env(self, saved_env):
        self.i8ctx.env = saved_env

    @property
    def wordsize(self):
        return self.i8ctx.wordsize

    @property
    def byteorder(self):
        return self.i8ctx.byteorder

    def implement(self, func, args, rets):
        provider, name = func.split("::")
        setattr(self, "%s_%s_impl" % (provider, name),
                StubImpl(weakref.ref(self), args, rets))

class StubImpl(object):
    def __init__(self, tref, args, rets):
        self.tref = tref
        self.args = args
        self.rets = rets

    def __call__(self, *args):
        self.tref().assertEqual(args, self.args)
        if len(self.rets) == 1:
            return self.rets[0]
        else:
            return self.rets



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from i8c.compiler import commands
from i8c import compiler
from i8c import runtime
from i8c import version
from i8c.compiler import target
from i8c.runtime.testcase import BaseTestCase
import io
import os
import struct
import subprocess

class SourceReader(io.BytesIO):
    def readline(self):
        line = io.BytesIO.readline(self)
        trim = line.find(b"//")
        if trim >= 0:
            line = line[:trim] + b"\n"
        return line

class TestOutput(runtime.Context):
    def __init__(self, testcase, index, asm):
        runtime.Context.__init__(self)
        self.__set_fileprefix(testcase, index)
        # Store the assembly language we generated
        asmfile = self.fileprefix + ".S"
        with open(asmfile, "wb") as fp:
            fp.write(asm)
        # Assemble it
        objfile = self.fileprefix + ".o"
        subprocess.check_call(
            commands.I8C_CC + ["-c", asmfile, "-o", objfile])
        # Load the notes from it
        self.import_notes(objfile)
        self.notes = []
        for notes in self.functions.values():
            self.notes.extend(notes)
        # Make sure we got at least one note
        testcase.assertGreaterEqual(len(self.notes), 1)
        # Setup for note execution
        self.env = testcase
        self.memory = testcase.memory
        self.memory.env = self
        self.register_symbol = testcase.register_symbol

    def __set_fileprefix(self, testcase, index):
        test_id = testcase.id().split(".")
        # Remove the common prefix and the name of the class
        assert test_id[0] == "tests"
        test_id.pop(0)
        test_id.pop(-2)
        # Build the result
        index = "_%04d" % index
        self.fileprefix = os.path.join(
            testcase.topdir, "tests", "output", *test_id) + index
        # Ensure the directory we'll write to exists
        dir = os.path.dirname(self.fileprefix)
        if not os.path.exists(dir):
            os.makedirs(dir)

    # See contrib/libi8x-testnote-export.py
    if os.environ.get("LIBI8X_TESTNOTE_EXPORT", "0") == "1":
        def __split_note_filename(self, filename):
            prefix = os.path.splitext(filename)[0]
            assert prefix == self.fileprefix
            prefix, testfunc = os.path.split(prefix)
            testfunc = testfunc.split("_")
            index = testfunc.pop()
            testfunc = "_".join(testfunc)
            prefix, pyfile = os.path.split(prefix)
            assert os.path.basename(prefix) == "output"
            return prefix, pyfile, testfunc, index

        def import_note(self, note):
            # First actually import the note
            runtime.Context.import_note(self, note)
            # Now decide where we'll save it
            prefix, pyfile, testfunc, index \
                = self.__split_note_filename(note.filename)
            self.export_count = getattr(self, "export_count", 0) + 1
            dir = os.path.join(
                prefix, "for-libi8x", "i8c", version(),
                "%d%s" % (self.wordsize,
                          {b"<": "el", b">": "be"}[self.byteorder]),
                pyfile, testfunc)
            filename = os.path.join(dir,
                                    "%s-%04d" % (index, self.export_count))
            # Now save it
            if not os.path.exists(dir):
                os.makedirs(dir)
            with open(filename, "wb") as fp:
                fp.write(note.bytes)

    @property
    def note(self):
        assert len(self.notes) == 1
        return self.notes[0]

    @property
    def ops(self):
        ops = sorted(self.note.ops.items())
        return [op for pc, op in ops]

    @property
    def opnames(self):
        return [op.name for op in self.ops]

class TestCase(BaseTestCase):
    _wordsize = target.guess_wordsize()

    def __locate_topdir(self):
        self.topdir = os.path.realpath(__file__)
        self.topdir, check = os.path.split(self.topdir)
        assert check.startswith("__init__.py")
        self.topdir, check = os.path.split(self.topdir)
        assert check == "tests"
        assert os.path.exists(os.path.join(self.topdir, "setup.py"))

    def run(self, *args, **kwargs):
        self.__locate_topdir()
        self.compilecount = 0
        return BaseTestCase.run(self, *args, **kwargs)

    def compile(self, input):
        self.compilecount += 1
        for line in input.split("\n"):
            if line.lstrip().startswith("wordsize "):
                break
        else:
            input = "wordsize %d\n%s" % (self._wordsize, input)
        input = SourceReader(b'# 1 "<testcase>"\n' + input.encode("utf-8"))
        output = io.BytesIO()
        tree = compiler.compile(input.readline, output.write)
        return tree, TestOutput(self, self.compilecount, output.getvalue())

    def disable_loggers(self):
        for logger in compiler.loggers.values():
            logger.disable()



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import RedefinedIdentError
from i8c.compiler import UndefinedIdentError

UNDEF_SOURCE = """\
define test::test_label
    goto label
"""

VALID_SOURCE = UNDEF_SOURCE + "label:\n"
REDEF_SOURCE = VALID_SOURCE + "label:\n"

class TestLabels(TestCase):
    """Check that undefined and duplicate labels are caught."""

    def test_valid(self):
        """Basic check of a valid label."""
        tree, output = self.compile(VALID_SOURCE)
        self.assertEqual([], output.opnames)

    def test_undefined(self):
        """Check that references to undefined labels are caught."""
        self.assertRaises(UndefinedIdentError, self.compile, UNDEF_SOURCE)

    def test_redefined(self):
        """Check that duplicate labels are inhibited."""
        self.assertRaises(RedefinedIdentError, self.compile, REDEF_SOURCE)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import ReservedIdentError

SOURCE = """\
define %s::test_reserved_provider returns ptr
    argument ptr __some_symbol
"""

class TestReservedProvider(TestCase):
    """Check that reserved provider names are rejected."""

    def test_reserved_provider(self):
        """Check that reserved provider names are rejected."""
        for provider in ("test", "libpthread", "i8test",
                         "i8core", "i8", "hello"):
            source = SOURCE % provider
            if provider.startswith("i8"):
                self.assertRaises(ReservedIdentError, self.compile, source)
            else:
                tree, output = self.compile(source)
                self.assertEqual([], output.opnames)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import cmdline
from ..compat import fprint
from . import blocks
from . import commands
from . import emitter
from . import externals
from . import I8CError
from . import lexer
from . import loggers
from . import names
from . import optimizer
from . import parser
from . import serializer
from . import stack
from . import target
from . import types
import copy
import io
import os
import subprocess
import sys

USAGE = """\
Usage: i8c [OPTION]... [FILE]...

Infinity Note Compiler.

Options:
  --help     Display this information.
  --version  Display version information.
  -E         Preprocess only; do not compile, assemble or link.
  -S         Compile only; do not assemble or link.
  -c         Compile and assemble, but do not link.
  -fpreprocessed
             Do not preprocess.
  --wrap-asm
             Wrap ‘-S’ output in C ‘asm’ statements.
  -o FILE    Place the output into FILE.

Note that I8C uses GCC both to preprocess its input (unless invoked
with ‘-fpreprocessed’) and to assemble its output (unless invoked with
‘-E’ or ‘-S’).  If GCC is used, all options not explicitly listed
above will be passed to GCC unmodified.

In general I8C operates like GCC, so if you’re used to GCC then I8C
should make sense.  Try it!

In most cases the command you want is ‘i8c -c file.i8’, which reads
and compiles ‘file.i8’ and writes the result to ‘file.o’.""" \
    + cmdline.usage_message_footer_for("I8C")

LICENSE = ("GPLv3+: GNU GPL version 3 or later",
           "http://gnu.org/licenses/gpl-3.0.html")

class CommandLine(object):
    def __init__(self, args):
        self.showinfo = None
        self.with_cpp = True
        self.with_i8c = True
        self.with_asm = True
        self.infiles = []
        self.outfile = None
        self.cpp_args = []
        self.asm_args = []
        self.wrap_asm = False
        self.__process_args(args)

    def __process_args(self, args):
        args = copy.copy(list(args))
        while args:
            arg = args.pop(0)

            # --help     Display usage information
            # --version  Display version information
            #
            # Both these options cause us to print information and
            # exit immediately, without continuing to process the
            # command line or compiling anything.
            if arg == "--help":
                self.showinfo = USAGE
                return

            elif arg == "--version":
                self.showinfo = cmdline.version_message_for("I8C", LICENSE)
                return

            # -E  Preprocess only; do not compile, assemble or link
            # -S  Compile only; do not assemble or link
            # -c  Compile and assemble, but do not link
            # -fpreprocessed
            #     Indicate to the preprocessor that the input file
            #     has already been preprocessed.
            #
            # These options control what processes we run.  GCC
            # doesn't seem to care if you specify more than one of
            # these so we don't either.
            elif arg == "-E":
                self.with_i8c = False
                self.with_asm = False

            elif arg == "-S":
                self.with_asm = False

            elif arg == "-c":
                self.asm_args.append(arg)

            elif arg == "-fpreprocessed":
                self.with_cpp = False

            elif arg == "--wrap-asm":
                self.wrap_asm = True

            # -o <file>  Place the output into <file>
            #
            # GCC doesn't complain about multiple "-o" options,
            # it just uses the last one it saw, so we do too.
            elif arg == "-o":
                if not args:
                    raise I8CError("missing filename after ‘-o’")

                self.asm_args.append(arg)
                self.outfile = args.pop(0)
                self.asm_args.append(self.outfile)

            elif arg.startswith("-o"):
                self.asm_args.append(arg)
                self.outfile = arg[2:]

            # Input filenames.  Not so easy to distinguish.
            elif (arg.endswith(".i8")
                  or arg.endswith(".i8p")) and not arg.startswith("-"):
                if self.asm_args and self.asm_args[-1] == "-include":
                    self.asm_args.pop()

                self.infiles.append(arg)
                self.cpp_args.append(arg)

            # -x <language>  Specify the language of input files
            #
            # Don't allow users to specify this, we need to use
            # it ourselves.
            elif arg.startswith("-x"):
                raise I8CError("unrecognized option ‘%s’" % arg)

            # --debug[=faculty1[,faculty2]...]
            #
            # Turn on debugging for some or all of i8c.
            elif arg.startswith("--debug"):
                if arg == "--debug":
                    for logger in loggers.values():
                        logger.enable()
                else:
                    for faculty in arg[8:].split(","):
                        logger = loggers.get(faculty, None)
                        if logger is not None:
                            logger.enable()

            # All other options get passed through to both the
            # preprocessor and the assembler, if they are used.
            else:
                self.cpp_args.append(arg)
                self.asm_args.append(arg)

def setup_input(args):
    process = infile = None
    if args.with_cpp:
        command = commands.I8C_CPP + ["-D__INFINITY__"] + args.cpp_args
        process = subprocess.Popen(command, stdout=subprocess.PIPE)
        infile = process.stdout
    elif args.infiles in ([], ["-"]):
        infile = sys.stdin
    else:
        infile = io.BytesIO()
        for filename in args.infiles:
            infile.write(open(filename, "rb").read())
        infile.seek(0)
    return process, infile

def guess_outfile(args):
    assert args.outfile is None
    if args.with_asm:
        assert "-c" in args.asm_args
        ext = ".o"
    elif args.wrap_asm:
        ext = ".c"
    else:
        ext = ".S"
    if len(args.infiles) != 1:
        raise I8CError("unable to determine output filename")
    root = os.path.splitext(args.infiles[0])[0]
    return root + ext

def setup_output(args):
    if args.with_asm:
        command = commands.I8C_AS + args.asm_args + ["-"]
        if args.outfile is None and "-c" in args.asm_args:
            command.extend(("-o", guess_outfile(args)))
        process = subprocess.Popen(command, stdin=subprocess.PIPE)
        outfile = process.stdin
    else:
        process = None
        filename = args.outfile
        if filename is None:
            if args.with_i8c:
                filename = guess_outfile(args)
            else:
                filename = "-"
        if filename == "-":
            outfile = sys.stdout
        else:
            outfile = open(filename, "wb")
    return process, outfile

def compile(readline, write, commandline=None):
    tree = parser.build_tree(lexer.generate_tokens(readline))
    tree.accept(target.TargetAnnotator(commandline))
    tree.accept(types.TypeAnnotator())
    tree.accept(names.NameAnnotator())
    tree.accept(externals.PerFileTableCreator())
    tree.accept(externals.PerFuncTableCreator())
    tree.accept(blocks.BlockCreator())
    tree.accept(stack.StackWalker())
    tree.accept(optimizer.BlockOptimizer())
    tree.accept(serializer.Serializer())
    tree.accept(optimizer.StreamOptimizer())
    tree.accept(emitter.Emitter(write, commandline))
    return tree

def main(args):
    args = CommandLine(args)
    if args.showinfo is not None:
        fprint(sys.stdout, args.showinfo)
        return

    clue = "Try ‘i8c --help’ for more information."
    if ((args.with_cpp and not args.cpp_args)
            or not (args.with_cpp or args.with_i8c or args.with_asm)):
        raise I8CError("nothing to do!\n%s" % clue)

    outfile = io.BytesIO()
    process, infile = setup_input(args)
    try:
        if args.with_i8c:
            compile(infile.readline, outfile.write, args)
        else:
            outfile.write(infile.read())

    finally:
        if process is not None:
            infile.close()
            process.wait()
            if process.returncode != 0:
                return process.returncode
    outfile.seek(0)
    infile = outfile

    process, outfile = setup_output(args)
    try:
        data = infile.read()
        if outfile is sys.stdout:
            data = data.decode("utf-8")
        outfile.write(data)
    finally:
        if process is not None:
            outfile.close()
            process.wait()
            if process.returncode != 0:
                return process.returncode



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import fwrite
import sys

class Logger(object):
    def __init__(self):
        self.is_enabled = False

    def disable(self):
        self.is_enabled = False

    def enable(self):
        self.is_enabled = True

    def __call__(self, msg):
        if self.is_enabled:
            sys.stderr.write(msg)

loggers = {}

def debug_printer_for(module):
    global loggers
    faculty = module.split(".")[-1]
    logger = loggers.get(faculty, None)
    if logger is None:
        logger = loggers[faculty] = Logger()
    return logger



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import I8CError, loggers
from i8c.compiler.driver import CommandLine

class TestCommandLineProcessor(TestCase):
    """Tests for the command line processor."""

    def setUp(self):
        self.disable_loggers()

    def tearDown(self):
        self.disable_loggers()

    def __process_command(self, cmd):
        return CommandLine(cmd.split())

    def test_empty(self):
        """Check with no arguments."""
        self.__check_empty(self.__process_command(""))

    def __check_empty(self, args):
        self.assertIs(args.showinfo, None)
        self.assertTrue(args.with_cpp)
        self.assertTrue(args.with_i8c)
        self.assertTrue(args.with_asm)
        self.assertEqual(args.infiles, [])
        self.assertIs(args.outfile, None)
        self.assertEqual(args.cpp_args, [])
        self.assertEqual(args.asm_args, [])

    def test_help(self):
        """Check that --help works."""
        for cmd in ("--help",
                    "--help --version",
                    "-O3 -E --help"):
            args = self.__process_command(cmd)
            self.assertIsNot(args.showinfo, None)
            self.assertGreaterEqual(
                args.showinfo.lower().find("usage:"), 0)

    def test_version(self):
        """Check that --version works."""
        for cmd in ("--version",
                    "--version --help",
                    "-O3 -E --version"):
            args = self.__process_command(cmd)
            self.assertIsNot(args.showinfo, None)
            self.assertGreaterEqual(
                args.showinfo.lower().find("no warranty"), 0)

    def test_E(self):
        """Check that -E works."""
        for cmd in ("-E",
                    "test.i8 -E -o test.i8i",
                    "-o test.o -S -E -fpreprocessed"):
            args = self.__process_command(cmd)
            self.assertIs(args.showinfo, None)
            self.assertFalse(args.with_i8c)
            self.assertFalse(args.with_asm)
            self.assertNotIn("-E", args.infiles)
            self.assertNotEqual(args.outfile, "-E")
            self.assertNotIn("-E", args.cpp_args)
            self.assertNotIn("-E", args.asm_args)

    def test_S(self):
        """Check that -S works."""
        for cmd in ("-S",
                    "test.i8 -S -o test.S",
                    "-o test.o -S -E -fpreprocessed"):
            args = self.__process_command(cmd)
            self.assertIs(args.showinfo, None)
            self.assertFalse(args.with_asm)
            self.assertNotIn("-S", args.infiles)
            self.assertNotEqual(args.outfile, "-S")
            self.assertNotIn("-S", args.cpp_args)
            self.assertNotIn("-S", args.asm_args)

    def test_c(self):
        """Check that -c works."""
        for cmd in ("-c",
                    "-c test.i8 -o test.S",
                    "test.i8 -c -o test.S",
                    "-o test.o -S -E -c -fpreprocessed"):
            args = self.__process_command(cmd)
            self.assertIs(args.showinfo, None)
            self.assertNotIn("-c", args.infiles)
            self.assertNotEqual(args.outfile, "-c")
            self.assertNotIn("-c", args.cpp_args)
            self.assertIn("-c", args.asm_args)

    def test_fpreprocessed(self):
        """Check that -fpreprocessed works."""
        for cmd in ("-fpreprocessed",
                    "-fpreprocessed test.i8 -o test.S",
                    "test.i8 -fpreprocessed -o test.S",
                    "-o test.o -S -E -c -fpreprocessed"):
            args = self.__process_command(cmd)
            self.assertIs(args.showinfo, None)
            self.assertFalse(args.with_cpp)
            self.assertNotIn("-fpreprocessed", args.infiles)
            self.assertNotEqual(args.outfile, "-fpreprocessed")
            self.assertNotIn("-fpreprocessed", args.cpp_args)
            self.assertNotIn("-fpreprocessed", args.asm_args)

    def test_o(self):
        """Check that -o works."""
        commands = ["-o",
                    "-fpreprocessed test.i8 -o",
                    "test.i8 -fpreprocessed -o"]
        for cmd in commands:
            self.assertRaises(I8CError, self.__process_command, cmd)

        cmd = "-o -S -E -c -fpreprocessed"
        args = self.__process_command(cmd)
        self.assertIs(args.showinfo, None)
        self.assertEqual(len(args.infiles), 0)
        self.assertEqual(args.outfile, "-S")
        self.assertNotIn("-o", args.cpp_args)
        self.assertNotIn("-S", args.cpp_args)
        self.assertIn("-o", args.asm_args)
        self.assertIn("-S", args.asm_args)

        commands.append(cmd)
        commands = [cmd.replace("-o", "-otest.o") for cmd in commands]
        self.__test_o(commands, "-otest.o")
        commands = [cmd.replace("-otest.o", "-o test.o") for cmd in commands]
        self.__test_o(commands, "-o", "test.o")

    def __test_o(self, commands, *expect):
        for cmd in commands:
            args = self.__process_command(cmd)
            self.assertIs(args.showinfo, None)
            self.assertEqual(args.outfile, "test.o")
            for arg in expect:
                self.assertNotIn(arg, args.cpp_args)
                self.assertIn(arg, args.asm_args)

    def test_x(self):
        """Check that -x is blocked."""
        commands = ("-x",
                    "-fpreprocessed -x",
                    "-x -fpreprocessed",
                    "-fpreprocessed test.i8 -x -o test.S",
                    "test.i8 -fpreprocessed -o test.S -x",
                    "-o test.o -S -E -x -c -fpreprocessed")
        self.__test_x(commands)
        commands = [cmd.replace("-x", "-xc") for cmd in commands]
        self.__test_x(commands)
        commands = [cmd.replace("-xc", "-x c") for cmd in commands]
        self.__test_x(commands)

    def __test_x(self, commands):
        for cmd in commands:
            self.assertRaises(I8CError, self.__process_command, cmd)

    def test_debug(self):
        """Check that --debug works."""
        args = self.__process_command("")
        self.__check_empty(args)
        for logger in loggers.values():
            self.assertFalse(logger.is_enabled)

        args = self.__process_command("--debug=serializer")
        self.__check_empty(args)
        for name, logger in loggers.items():
            if name == "serializer":
                self.assertTrue(logger.is_enabled)
            else:
                self.assertFalse(logger.is_enabled)

        args = self.__process_command("--debug")
        self.__check_empty(args)
        for logger in loggers.values():
            self.assertTrue(logger.is_enabled)

    def test_include(self):
        """Check that -include works."""
        for cmd in ("-include test.i8",
                    "-c -include test.i8",
                    "-include test.i8 -c"):
            args = self.__process_command(cmd)
            self.assertIn("test.i8", args.infiles)
            self.assertNotEqual(args.outfile, "test.i8")
            for arg in ("-include", "test.i8"):
                self.assertIn(arg, args.cpp_args)
                self.assertNotIn(arg, args.asm_args)

    def test_input_files(self):
        """Check that input files are handled correctly."""
        for cmd in ("test.i8",
                    "-fpreprocessed test.i8",
                    "test.i8 -fpreprocessed",
                    "-fpreprocessed -o test.S test.i8",
                    "test.i8 -fpreprocessed -o test.S",
                    "-o test.o -S -E -c  test.i8 -fpreprocessed"):
            args = self.__process_command(cmd)
            self.assertIn("test.i8", args.infiles)
            self.assertNotEqual(args.outfile, "test.i8")
            self.assertIn("test.i8", args.cpp_args)
            self.assertNotIn("test.i8", args.asm_args)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import blocks
from . import logger
from . import operations
from . import visitors

debug_print = logger.debug_printer_for(__name__)

class StopBlock(blocks.BasicBlock):
    def __init__(self, ast):
        blocks.BasicBlock.__init__(self, None)
        self.append(StopOp(ast))

class StopOp(operations.Operation):
    @property
    def source(self):
        return "EOF"

class OperationStream(visitors.Visitable):
    def __init__(self, function):
        self.ops = {}
        self.jumps = {}
        self.labels = {}
        self.__eof_block = StopBlock(function)

    @property
    def is_closed(self):
        return self.__eof_block is None

    @property
    def items(self):
        assert self.is_closed
        return self.ops.items

    @property
    def stream(self):
        return sorted(self.items())

    # Methods used to build the stream

    @property
    def last_op(self):
        assert not self.is_closed
        assert self.ops
        return self.ops[len(self.ops) - 1]

    def append(self, op):
        assert not self.is_closed
        self.ops[len(self.ops)] = op

    def jump_from_last_to(self, target):
        assert not self.is_closed
        source, target = self.last_op, target.first_op

        if target not in self.labels:
            self.labels[target] = []
        self.labels[target].append(source)

        assert source not in self.jumps
        self.jumps[source] = target

    def jump_from_last_to_EOF(self):
        self.jump_from_last_to(self.__eof_block)

    def close(self):
        assert not self.is_closed
        self.append(self.__eof_block.first_op)
        self.__eof_block = None
        self.__replace_returns_with_gotos()

    def __replace_returns_with_gotos(self):
        for index, op in self.ops.items():
            if not op.is_return:
                continue
            self.replace_by_index(index, operations.SyntheticGoto(op))

    # Methods used to mutate the stream

    def replace_by_index(self, index, new_op):
        self.replace_by_index_op(index, self.ops[index], new_op)

    def replace_by_index_op(self, index, old_op, new_op):
        assert self.is_closed

        for dict in self.labels, self.jumps:
            assert new_op not in dict
            tmp = dict.pop(old_op, None)
            if tmp is not None:
                dict[new_op] = tmp

        for lbl_ops in self.labels.values():
            for lbl_op, lbl_index in zip(lbl_ops, range(len(lbl_ops))):
                if lbl_op is old_op:
                    lbl_ops[lbl_index] = new_op

        for jump_src, jump_tgt in self.jumps.items():
            if jump_tgt is old_op:
                self.jumps[jump_src] = new_op

        self.ops[index] = new_op

    def remove_by_index_op(self, index, op):
        assert self.is_closed
        assert op not in self.labels

        target = self.jumps.pop(op, None)
        if target is not None:
            self.__remove_label(target, op)

        check = self.ops.pop(index)
        assert op == check
        self.__reindex()

    def retarget_jump(self, from_op, to_op):
        assert self.is_closed

        trampoline = self.jumps[from_op]
        self.__remove_label(trampoline, from_op)
        self.jumps[from_op] = to_op
        self.labels[to_op].append(from_op)

    def __remove_label(self, target, source):
        assert self.is_closed
        self.labels[target].remove(source)
        if not self.labels[target]:
            del self.labels[target]

    def __reindex(self):
        ops = self.stream
        self.ops = {}
        for index, op in ops:
            self.ops[len(self.ops)] = op

    # Debug printing

    def __str__(self):
        reverse = {}
        for index, op in self.ops.items():
            reverse[op] = index
        lines = []
        for index, op in self.ops.items():
            line = []
            sources = self.labels.get(op, None)
            if sources is not None:
                line.append("%d:" % index)
            line.append("\t%s" % op)
            jump = self.jumps.get(op, None)
            if jump is not None:
                line.append(" -> %d" % reverse[jump])
            lines.append("".join(line))
        return "\n".join(lines)

class Serializer(object):
    def visit_toplevel(self, toplevel):
        for node in toplevel.functions:
            node.accept(self)

    def visit_function(self, function):
        self.stream = OperationStream(function)
        self.visited = {}
        function.entry_block.accept(self)
        del function.entry_block
        self.stream.close()
        debug_print("%s:\n" % function.name.value)
        debug_print("%s\n\n" % self.stream)
        function.ops = self.stream

    def visit_basicblock(self, block):
        if self.visited.get(block, False):
            return
        self.visited[block] = True

        for op in block.ops:
            self.stream.append(op)

        if block.is_branch_terminated:
            self.stream.jump_from_last_to(block.branched_exit)

            self.stream.append(operations.SyntheticGoto(self.stream.last_op))
            self.stream.jump_from_last_to(block.nobranch_exit)

            block.nobranch_exit.accept(self)
            block.branched_exit.accept(self)
        elif block.is_goto_terminated:
            self.stream.jump_from_last_to(block.goto_exit)

            block.goto_exit.accept(self)
        else:
            assert block.is_return_terminated
            assert not block.exits

            self.stream.jump_from_last_to_EOF()



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

SOURCE = """\
define test::stack_ops_test
    argument int arg_1
    argument ptr arg_2
    argument bool arg_3

    dup
    drop
    pick 0
    pick 1
    pick 2
    over
    swap
    rot
"""

class TestStackOperations(TestCase):
    def test_stack_ops(self):
        """Basic checks for stack-manipulation bytecodes."""
        tree, output = self.compile(SOURCE)
        # Check the assembler contains the expected operations
        self.assertEqual(["dup",
                          "drop",
                          "dup",    # pick 0
                          "over",   # pick 1
                          "pick",   # pick 2
                          "over",
                          "swap",
                          "rot"], output.opnames)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import archspec
from .. import constants
from ..compat import fwrite, str
from . import logger
from . import types
from . import visitors
from .types import PTRTYPE

debug_print = logger.debug_printer_for(__name__)

class Label(object):
    def __init__(self, name):
        self.name = name
        self.emitted = False

    @property
    def ref(self):
        return self.name + (self.emitted and "b" or "f")

    def __sub__(self, other):
        if other is self:
            return "0"
        else:
            return "%s-%s" % (self.ref, other.ref)

class String(object):
    def __init__(self, table, text):
        self.text = text

    def append(self, more):
        self.text += more

    @staticmethod
    def quote(text):
        return '"%s"' % text.replace("\\", "\\\\").replace('"', '\\"')

    @property
    def quoted(self):
        return String.quote(self.text)

class StringTable(object):
    def __init__(self):
        self.strings = []
        self.entries = []

    @property
    def is_laid_out(self):
        return len(self.entries) != 0

    @property
    def start_label(self):
        return self.entries[0][0]

    def new(self, text=""):
        assert not self.is_laid_out
        result = String(self, text)
        self.strings.append(result)
        return result

    def layout_table(self, new_label):
        assert not self.is_laid_out
        # Create a length-sorted list of unique strings
        unique = {}
        for string in self.strings:
            unique[string.text] = True
        unique = [(-len(text), text) for text in unique.keys()]
        unique.sort()
        # Calculate offsets for each
        offsets = {}
        for junk, text in unique:
            for label, entry_text in self.entries:
                if entry_text.endswith(text):
                    offsets[text] = "%s+%d" % (
                        offsets[entry_text],
                        len(entry_text) - len(text))
                    break
            else:
                label = new_label()
                self.entries.append((label, text))
                offsets[text] = label - self.start_label
        # Store the offsets in the strings
        for string in self.strings:
            string.offset = offsets[string.text]

    def emit(self, emitter):
        for label, text in self.entries:
            emitter.emit_label(label)
            emitter.emit(".string " + String.quote(text))

class ExternTable(object):
    def __init__(self, funcname, strings):
        self.strings = strings
        self.indexes = {str(funcname): 0}
        self.entries = []

    def index_of(self, external):
        key = str(external.name)
        index = self.indexes.get(key, None)
        if index is not None:
            return index
        index = len(self.entries) + 1
        type = external.basetype
        assert type.is_function
        self.entries.append(FuncRef(*map(self.strings.new, (
            external.name.provider,
            external.name.name,
            "".join((t.encoding for t in type.paramtypes)),
            "".join((t.encoding for t in type.returntypes))))))
        self.indexes[key] = index
        return index

    def emit(self, emitter):
        for entry, index in zip(self.entries, range(1, len(self.entries) + 1)):
            entry.emit(emitter, "extern %d " % index)

class FuncRef(object):
    def __init__(self, provider, name, params, returns):
        self.provider = provider
        self.name = name
        self.params = params
        self.returns = returns

    def emit(self, emitter, prefix):
        emitter.emit_uleb128(self.provider.offset, prefix + "provider offset")
        emitter.emit_uleb128(self.name.offset, prefix + "name offset")
        emitter.emit_uleb128(self.params.offset, prefix + "ptypes offset")
        emitter.emit_uleb128(self.returns.offset, prefix + "rtypes offset")

class TableBuilder(object):
    def __init__(self, emitter):
        self.emitter = emitter

    def visit_function(self, function):
        funcname = function.name.value
        assert funcname.is_fullname
        self.strings = StringTable()
        self.externs = ExternTable(funcname, self.strings)

        # Create strings for the signature chunk.
        self.emitter.provider = self.strings.new(funcname.provider)
        self.emitter.name = self.strings.new(funcname.name)
        self.emitter.paramtypes = self.strings.new()
        function.parameters.accept(self)
        self.emitter.returntypes = self.strings.new()
        function.returntypes.accept(self)

        # Create externals and strings for the bytecode chunk.
        function.ops.accept(self)

        # Lay out the strings table.
        self.strings.layout_table(self.emitter.new_label)

        self.emitter.strings = self.strings
        self.emitter.externs = self.externs

    def visit_parameters(self, parameters):
        for node in parameters.children:
            node.accept(self)

    def visit_parameter(self, param):
        self.emitter.paramtypes.append(param.typename.type.encoding)

    def visit_returntypes(self, returntypes):
        for node in returntypes.children:
            self.emitter.returntypes.append(node.type.encoding)

    def visit_operationstream(self, ops):
        for index, op in ops.stream:
            op.accept(self)

    def visit_loadop(self, op):
        if op.is_loadext and op.external.type.is_function:
            op.etable_index = self.externs.index_of(op.external)

    def visit_warnop(self, op):
        op.string = self.strings.new(op.value)

    def visit_operation(self, op):
        pass

class NoOutputOpSkipper(object):
    def visit_nameop(self, op):
        pass

    def visit_stopop(self, op):
        pass

class Emitter(NoOutputOpSkipper):
    def __init__(self, write, commandline=None):
        self.__write = write
        self.wrap_asm = commandline is not None and commandline.wrap_asm

    def write(self, text):
        self.__write(text.encode("utf-8"))

    def write_asm(self, line, comment=None):
        if self.wrap_asm and line:
            line = r'  "%s\n"' % line.replace('"', r'\"')
        if comment is not None:
            line += "\t/* %s */" % comment
        line += "\n"
        self.write(line)

    def new_label(self):
        self.num_labels += 1
        return Label(str(self.num_labels))

    def emit(self, line, comment=None):
        if line.startswith("."):
            line = "\t" + line
        if not line.startswith("#") and self.__label is not None:
            line = "%s:%s" % (self.__label.name, line)
            self.__label = None
        self.write_asm(line, comment)

    def emit_newline(self):
        self.emit("")

    def emit_comment(self, comment):
        self.emit("", comment)

    def emit_label(self, label):
        if self.__label is not None:
            self.write_asm("%s:" % self.__label.name)
            self.__label = None
        self.__label = label
        label.emitted = True

    def __emit_constant(self, directive, value, comment):
        value = str(value)
        tmp = getattr(constants, value, None)
        if tmp is not None:
            if comment is None:
                comment = value
            else:
                comment = "%s (%s)" % (value, comment)
            if value.startswith("I8_OP_"):
                tmp -= 0x100
            value = tmp
        self.emit(".%s %s" % (directive, value), comment)

    def __emit_nbyte(self, n, value, comment):
        self.__emit_constant(("%sbyte" % n), value, comment)

    def emit_byte(self, value, comment=None):
        self.__emit_nbyte("", value, comment)

    def emit_2byte(self, value, comment=None):
        self.__emit_nbyte(2, value, comment)

    def emit_4byte(self, value, comment=None):
        self.__emit_nbyte(4, value, comment)

    def emit_8byte(self, value, comment=None):
        self.__emit_nbyte(8, value, comment)

    def emit_uleb128(self, value, comment=None):
        self.__emit_constant("uleb128", value, comment)

    def emit_sleb128(self, value, comment=None):
        self.__emit_constant("sleb128", value, comment)

    def emit_op(self, name, comment=None):
        widename = "I8_OP_" + name
        widecode = getattr(constants, widename, None)
        if widecode is not None:
            name = "GNU_wide_op"
        name = "DW_OP_" + name
        self.emit_byte(name, comment)
        if widecode is not None:
            self.emit_uleb128(widename)

    def visit_toplevel(self, toplevel):
        self.num_labels = 0
        self.__label = None
        self.wordsize = toplevel.wordsize
        bytes, check = divmod(self.wordsize, 8)
        assert check == 0
        self.emit_address = getattr(self, "emit_%dbyte" % bytes)
        if self.wrap_asm:
            self.write("__asm__ (\n")
        self.emit('.pushsection .note.infinity, "", "note"')
        self.emit(".balign 4")
        for node in toplevel.functions:
            node.accept(self)
        self.emit(".popsection")
        if self.wrap_asm:
            self.write("  );\n")

    def visit_function(self, function):
        debug_print("\n%s:\n%s\n" % (function.name.value, function.ops))
        # This method handles laying out the structure of the note
        # as per http://www.netbsd.org/docs/kernel/elf-notes.html.
        # The Infinity-specific part is in the "desc" field and is
        # handled by self.__visit_function.
        namestart = self.new_label()
        namelimit = self.new_label()
        descstart = self.new_label()
        desclimit = self.new_label()
        self.emit_newline()
        self.emit_comment(function.name.ident)
        self.emit_4byte(namelimit - namestart, "namesz")
        self.emit_4byte(desclimit - descstart, "descsz")
        self.emit_4byte("NT_GNU_INFINITY")
        self.emit_label(namestart)
        self.emit('.string "GNU"')
        self.emit_label(namelimit)
        self.emit(".balign 4")
        self.emit_label(descstart)
        self.__visit_function(function)
        self.emit_label(desclimit)
        self.emit(".balign 4")

    def __visit_function(self, function):
        # Build the strings and externals tables.
        function.accept(TableBuilder(self))

        # Emit the code chunks if required.
        if self.has_code(function):
            self.emit_chunk("codeinfo", 1, Emitter.emit_codeinfo, function)
            self.emit_chunk("bytecode", 3, Emitter.emit_bytecode, function)

        # Emit the remaining chunks.
        self.emit_chunk("signature", 2, Emitter.emit_signature)
        if self.externs.entries:
            self.emit_chunk("externals", 2, self.externs.emit)
        self.emit_chunk("strings", 1, self.strings.emit)

    def emit_chunk(self, name, version, emitfunc, *args):
        start = self.new_label()
        limit = self.new_label()
        self.emit_uleb128("I8_CHUNK_" + name.upper())
        self.emit_uleb128(version, "chunk version")
        self.emit_uleb128(limit - start, "chunk size")
        self.emit_label(start)
        emitfunc(self, *args)
        self.emit_label(limit)

    def emit_signature(self):
        self.emit_uleb128(self.provider.offset, "provider offset")
        self.emit_uleb128(self.name.offset, "name offset")
        self.emit_uleb128(self.paramtypes.offset, "param types offset")
        self.emit_uleb128(self.returntypes.offset, "return types offset")

    @staticmethod
    def has_code(function):
        skipper = NoOutputOpSkipper()
        try:
            for index, op in function.ops.stream:
                op.accept(skipper)
            return False
        except visitors.VisitError:
            return True

    def emit_codeinfo(self, function):
        self.emit_2byte(archspec.encode(self.wordsize), "archspec")
        self.emit_uleb128(function.max_stack, "max stack")

    def emit_bytecode(self, function):
        function.ops.accept(self)

    # Emit the bytecode

    def visit_operationstream(self, ops):
        self.jumps = ops.jumps
        self.labels = {}
        for op in ops.labels.keys():
            self.labels[op] = self.new_label()
        for index, op in ops.stream:
            label = self.labels.get(op, None)
            if label is not None:
                self.emit_label(label)
            op.accept(self)

    # Generic visitors that handle groups of operations.

    def visit_nooperandsop(self, op):
        self.emit_op(op.dwarfname, op.fileline)

    def visit_terminalop(self, op):
        assert op.is_branch or op.is_goto
        target = self.labels[self.jumps[op]]
        source = self.new_label()
        self.emit_op(op.dwarfname, op.fileline)
        self.emit_2byte(target - source)
        self.emit_label(source)

    # Visitors for operations that need specific handling.

    def visit_castop(self, op):
        self.emit_op("cast_%s2%s" % (op.old_type.basetype.name,
                                     op.new_type.basetype.name))
        self.emit_uleb128(op.slot)

    def visit_constop(self, op):
        value = op.value
        if value >= 0:
            if value < 0x20:
                self.emit_op("lit%d" % value, op.fileline)
            elif value < (1 << 8):
                self.emit_op("const1u", op.fileline)
                self.emit_byte(value)
            elif value < (1 << 16):
                self.emit_op("const2u", op.fileline)
                self.emit_2byte(value)
            elif value < (1 << 21):
                self.emit_op("constu", op.fileline)
                self.emit_uleb128(value)
            elif value < (1 << 32):
                self.emit_op("const4u", op.fileline)
                self.emit_4byte(value)
            elif value < (1 << 49):
                self.emit_op("constu", op.fileline)
                self.emit_uleb128(value)
            elif value < (1 << 64):
                self.emit_op("const8u", op.fileline)
                self.emit_8byte(value)
            else:
                self.emit_op("constu", op.fileline)
                self.emit_uleb128(value)
        else:
            if value >= -(1 << 7):
                self.emit_op("const1s", op.fileline)
                self.emit_byte(value)
            elif value >= -(1 << 15):
                self.emit_op("const2s", op.fileline)
                self.emit_2byte(value)
            elif value >= -(1 << 20):
                self.emit_op("consts", op.fileline)
                self.emit_sleb128(value)
            elif value >= -(1 << 31):
                self.emit_op("const4s", op.fileline)
                self.emit_4byte(value)
            elif value >= -(1 << 48):
                self.emit_op("consts", op.fileline)
                self.emit_sleb128(value)
            elif value >= -(1 << 63):
                self.emit_op("const8s", op.fileline)
                self.emit_8byte(value)
            else:
                self.emit_op("consts", op.fileline)
                self.emit_sleb128(value)

    def visit_derefop(self, op):
        if op.type.basetype is types.PTRTYPE:
            self.emit_op("deref", op.fileline)
            return

        sizedtype = op.type.sizedtype
        operand = sizedtype.nbits
        if operand is None:
            operand = self.wordsize
        if sizedtype.is_signed:
            operand *= -1

        # "deref_int 0" used to mean a word-sized integer, but its
        # signedness was ambiguous so this use has been deprecated.
        assert operand != 0

        self.emit_op("deref_int", op.fileline)
        self.emit_sleb128(operand)

    def visit_loadop(self, op):
        self.emit_op(op.dwarfname, op.fileline)
        if op.is_pick:
            if op.pickslot > 1:
                self.emit_byte(op.pickslot)
        else:
            assert op.is_loadext
            if op.external.basetype is PTRTYPE:
                self.emit_address(op.external.name)
            else:
                assert op.external.type.is_function
                self.emit_uleb128(op.etable_index)

    def visit_plusuconst(self, op):
        self.emit_op("plus_uconst", op.fileline)
        self.emit_uleb128(op.value)

    def visit_warnop(self, op):
        self.emit_op("warn", op.fileline)
        self.emit_uleb128(op.string.offset, op.string.quoted)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import main
import sys

class TestEntryPoint(TestCase):
    """Test i8c.compiler.main, the console scripts entry point.

    This testcase should be the bare minimum required to exercise
    i8c.compiler.main.  Tests exercising the function it wraps
    (i8c.compiler.driver.main) should be in test_compiler_driver.py
    so they may be run without messing with sys.argv and sys.stderr.
    """

    def setUp(self):
        self.saved_argv = sys.argv
        self.saved_stderr = sys.stderr

    def tearDown(self):
        sys.argv = self.saved_argv
        sys.stderr = self.saved_stderr

    def test_success_path(self):
        """Check the i8c console scripts entry point success path."""
        sys.argv[1:] = ["--version"]
        self.assertIs(main(), None)

    def test_failure_path(self):
        """Check the i8c console scripts entry point failure path."""
        sys.argv[1:] = ["-x"]
        sys.stderr = sys.stdout
        self.assertEqual(main(), 1)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import StackError

# All binary operations were considered equal to each other: the "mul"
# and "add" were considered equivalent and the branch optimized away.

SOURCE = """\
define test::i8cfail_0003 returns int
    argument int y
    argument int z

    load 0
    beq label

    load 0
    mul
    return

label:
    load 1
    add
"""

class TestI8CFail0003(TestCase):
    def test_i8cfail_0003(self):
        """Miscellaneous I8C failure #0003 check"""
        tree, output = self.compile(SOURCE)
        ops = output.opnames
        # Robust test: check the stream contains a branch
        self.assertIn("bra", ops)
        # Fragile test: check that the stream is as we expect.
        # This may need fixing up to cope with future compiler changes.
        self.assertEqual(["bra",
                          "plus_uconst", "skip",
                          "lit0", "mul"], ops)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import StackError

SOURCE = """\
define test::unary_ops_test
    argument %s arg
    %s
"""

class TestUnaryOperations(TestCase):
    OPERATIONS = "abs", "neg", "not"

    def test_int_arg(self):
        """Check that unary operations work with int arguments."""
        for type in ("int", "bool"):
            for op in self.OPERATIONS:
                tree, output = self.compile(SOURCE % (type, op))
                self.assertEqual([op], output.opnames)

    def test_nonint_arg(self):
        """Check unary operations with non-int arguments are rejected."""
        for type in ("ptr", "opaque", "func ()"):
            for op in self.OPERATIONS:
                self.assertRaises(StackError, self.compile,
                                  SOURCE % (type, op))



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import logger
from . import ParserError
from . import ReservedIdentError

debug_print = logger.debug_printer_for(__name__)

class Name(object):
    is_builtin = False

    def __init__(self, ast, provider, name):
        assert provider is None or provider
        assert name
        self.ast = ast
        self.provider = provider
        self.name = name

    @property
    def fileline(self):
        return self.ast.fileline

    @property
    def is_shortname(self):
        return self.provider is None

    @property
    def is_fullname(self):
        return not self.is_shortname

    def with_provider(self, provider):
        assert self.is_shortname
        return Name(self.ast, provider, self.name)

    def without_provider(self):
        assert self.is_fullname
        return Name(self.ast, None, self.name)

    def __eq__(self, other): # pragma: no cover
        # This comparison is excluded from coverage because it's
        # not currently entered (but it must be defined because
        # we've defined __ne__ below).
        return not (self != other)

    def __ne__(self, other):
        return (other is None
                or self.provider != other.provider
                or self.name != other.name)

    def __str__(self):
        if self.provider is not None:
            return "%s::%s" % (self.provider, self.name)
        else:
            return self.name

class NameAnnotator(object):
    def visit_toplevel(self, toplevel):
        for node in toplevel.children:
            self.function = None
            node.accept(self)

    def visit_wordsize(self, wordsize):
        pass

    def visit_typedef(self, typedef):
        pass

    def visit_function(self, function):
        if function.provider.startswith("i8"):
            raise ReservedIdentError(
                function, "provider", function.provider)
        self.function = function

        self.allow_shortname = False
        self.allow_fullname = True
        for node in function.children:
            node.accept(self)

        if debug_print.is_enabled:
            debug_print("%s\n\n" % function)

    def visit_parameters(self, parameters):
        self.allow_shortname = True
        self.allow_fullname = False

        for node in parameters.children:
            node.accept(self)

    def visit_parameter(self, parameter):
        parameter.name.accept(self)

    def visit_externals(self, externals):
        for node in externals.children:
            node.accept(self)

    def visit_external(self, external):
        if external.typename.type.is_function:
            self.allow_shortname = False
            self.allow_fullname = True
        else:
            self.allow_shortname = True
            self.allow_fullname = False

        external.name.accept(self)

    def visit_returntypes(self, returntypes):
        pass

    def visit_operations(self, ops):
        self.allow_shortname = True
        self.allow_fullname = True

        for node in ops.children:
            node.accept(self)

    def visit_label(self, label):
        pass

    def visit_operation(self, op):
        pass

    def visit_loadop(self, op):
        for node in op.named_operands:
            node.accept(self)

    def visit_namecastop(self, op):
        for node in op.named_operands:
            node.accept(self)

    # Visitors for names

    def visit_fullname(self, node):
        if not self.allow_fullname:
            raise ParserError(node.tokens[1:])
        node.value = Name(node, node.provider, node.shortname)

    def visit_shortname(self, node):
        if self.allow_shortname:
            provider = None
        elif self.allow_fullname and self.function is not None:
            provider = self.function.provider
        else:
            raise ParserError(node.tokens)
        node.value = Name(node, provider, node.name)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import constants
from . import BadDerefError
from . import UnhandledNoteError
from . import leb128
import operator
import struct

class Operation(object):
    NAMES = {}
    for name in dir(constants):
        if name[2:6] == "_OP_":
            assert name not in NAMES
            NAMES[getattr(constants, name)] = name
    del name

    OPERANDS = {
        constants.DW_OP_addr: ["address"],
        constants.DW_OP_const1u: ["u1"],
        constants.DW_OP_const1s: ["s1"],
        constants.DW_OP_const2u: ["u2"],
        constants.DW_OP_const2s: ["s2"],
        constants.DW_OP_const4u: ["u4"],
        constants.DW_OP_const4s: ["s4"],
        constants.DW_OP_const8u: ["u8"],
        constants.DW_OP_const8s: ["s8"],
        constants.DW_OP_constu: ["uleb128"],
        constants.DW_OP_consts: ["sleb128"],
        constants.DW_OP_pick: ["u1"],
        constants.DW_OP_plus_uconst: ["uleb128"],
        constants.DW_OP_bra: ["s2"],
        constants.DW_OP_skip: ["s2"],
        constants.DW_OP_breg0: ["sleb128"],
        constants.DW_OP_breg1: ["sleb128"],
        constants.DW_OP_breg2: ["sleb128"],
        constants.DW_OP_breg3: ["sleb128"],
        constants.DW_OP_breg4: ["sleb128"],
        constants.DW_OP_breg5: ["sleb128"],
        constants.DW_OP_breg6: ["sleb128"],
        constants.DW_OP_breg7: ["sleb128"],
        constants.DW_OP_breg8: ["sleb128"],
        constants.DW_OP_breg9: ["sleb128"],
        constants.DW_OP_breg10: ["sleb128"],
        constants.DW_OP_breg11: ["sleb128"],
        constants.DW_OP_breg12: ["sleb128"],
        constants.DW_OP_breg13: ["sleb128"],
        constants.DW_OP_breg14: ["sleb128"],
        constants.DW_OP_breg15: ["sleb128"],
        constants.DW_OP_breg16: ["sleb128"],
        constants.DW_OP_breg17: ["sleb128"],
        constants.DW_OP_breg18: ["sleb128"],
        constants.DW_OP_breg19: ["sleb128"],
        constants.DW_OP_breg20: ["sleb128"],
        constants.DW_OP_breg21: ["sleb128"],
        constants.DW_OP_breg22: ["sleb128"],
        constants.DW_OP_breg23: ["sleb128"],
        constants.DW_OP_breg24: ["sleb128"],
        constants.DW_OP_breg25: ["sleb128"],
        constants.DW_OP_breg26: ["sleb128"],
        constants.DW_OP_breg27: ["sleb128"],
        constants.DW_OP_breg28: ["sleb128"],
        constants.DW_OP_breg29: ["sleb128"],
        constants.DW_OP_breg30: ["sleb128"],
        constants.DW_OP_breg31: ["sleb128"],
        constants.DW_OP_regx: ["uleb128"],
        constants.DW_OP_fbreg: ["sleb128"],
        constants.DW_OP_bregx: ["uleb128", "sleb128"],
        constants.DW_OP_piece: ["uleb128"],
        constants.DW_OP_deref_size: ["u1"],
        constants.DW_OP_xderef_size: ["u1"],
        constants.I8_OP_load_external: ["uleb128"],
        constants.I8_OP_deref_int: ["sleb128"],
        constants.I8_OP_cast_int2ptr: ["uleb128"],
        constants.I8_OP_cast_ptr2int: ["uleb128"],
        constants.I8_OP_warn: ["string"],
    }

    OPTABLE = {
        constants.DW_OP_abs: (abs, 1, True),
        constants.DW_OP_and: (operator.and_, 2, False),
        constants.DW_OP_div: (operator.floordiv, 2, True),
        constants.DW_OP_minus: (operator.sub, 2, False),
        constants.DW_OP_mod: (operator.mod, 2, False),
        constants.DW_OP_mul: (operator.mul, 2, False),
        constants.DW_OP_or: (operator.or_, 2, False),
        constants.DW_OP_neg: (operator.neg, 1, True),
        constants.DW_OP_not: (operator.invert, 1, False),
        constants.DW_OP_plus: (operator.add, 2, False),
        constants.DW_OP_shl: (operator.lshift, 2, False),
        constants.DW_OP_shr: (operator.rshift, 2, False),
        constants.DW_OP_shra: (operator.rshift, 2, True),
        constants.DW_OP_xor: (operator.xor, 2, False),
        constants.DW_OP_eq: (operator.eq, 2, True),
        constants.DW_OP_ge: (operator.ge, 2, True),
        constants.DW_OP_gt: (operator.gt, 2, True),
        constants.DW_OP_le: (operator.le, 2, True),
        constants.DW_OP_lt: (operator.lt, 2, True),
        constants.DW_OP_ne: (operator.ne, 2, True),
    }

    FIXEDSIZE = {}
    for code in "bBhHiIqQ":
        code = bytes(code.encode("utf-8"))
        size = struct.calcsize(code)
        type = "%s%d" % (code.isupper() and "u" or "s", size)
        assert type not in FIXEDSIZE
        FIXEDSIZE[type] = size, code
    del code, size, type

    def __init__(self, function, pc):
        src = function.bytecode + pc
        self.get_string = function.get_string
        # Read the opcode
        self.opcode = ord(src[0])
        next = src + 1
        if self.opcode == constants.DW_OP_GNU_wide_op:
            size, widecode = self.decode_uleb128(next)
            self.opcode = widecode + 0x100
            next += size
        if self.opcode not in self.NAMES:
            raise UnhandledNoteError(src)
        # Read the operands
        self.operands = []
        for type in self.OPERANDS.get(self.opcode, ()):
            sizecode = self.FIXEDSIZE.get(type, None)
            if sizecode is not None:
                size, fmt = sizecode
                fmt = src.byteorder + fmt
                value = struct.unpack(fmt, next[:size].bytes)[0]
            else:
                size, value = getattr(self, "decode_" + type)(next)
            self.operands.append(value)
            next += size
        # Store our source location for exceptions
        self.src = src[:next.start - src.start]
        # Store our location and encoded form for tracing
        self.location = (function, pc)
        self.encoded = self.src.text
        # Counter for coverage checks
        self.hitcount = 0
        # Tidy up
        del self.get_string

    @property
    def size(self):
        return len(self.src)

    @property
    def byteorder(self):
        return self.src.byteorder

    @property
    def function(self):
        return self.location[0]

    @staticmethod
    def decode_address(code):
        return code.wordsize >> 3, code.symbol_names

    @staticmethod
    def decode_uleb128(code):
        return leb128.read_uleb128(code, 0)

    @staticmethod
    def decode_sleb128(code):
        return leb128.read_sleb128(code, 0)

    def decode_string(self, code):
        size, index = self.decode_uleb128(code)
        return size, self.get_string(index).text

    @property
    def name(self):
        result = self.NAMES[self.opcode]
        assert result[2:6] == "_OP_"
        return result[6:]

    @property
    def operand(self):
        assert len(self.operands) == 1
        return self.operands[0]

    def __trace(self, ctx, stack):
        ctx.trace_operation(self.location, stack,
                            " ".join("%02x" % ord(c)
                                     for c in self.encoded),
                            " ".join([self.NAMES[self.opcode]]
                                     + list(map(repr, self.operands))))

    def execute(self, ctx, externals, stack):
        self.__trace(ctx, stack)
        self.hitcount += 1
        if (self.opcode >= constants.DW_OP_lit0
              and self.opcode <= constants.DW_OP_lit31):
            impl = self.__exec_litN
        elif (self.opcode >= constants.DW_OP_const1u
              and self.opcode <= constants.DW_OP_consts):
            impl = self.__exec_constX
        elif self.opcode in self.OPTABLE:
            impl = self.__exec_optable
        else:
            impl = getattr(self, "exec_" + self.name, None)
        if impl is None:
            raise NotImplementedError(self.name)
        return impl(ctx, externals, stack)

    def __exec_optable(self, ctx, externals, stack):
        func, num_args, is_signed = self.OPTABLE[self.opcode]
        pop = is_signed and stack.pop_signed or stack.pop_unsigned
        if num_args == 2:
            impl = self.__exec_binary
        else:
            assert num_args == 1
            impl = self.__exec_unary
        return impl(ctx, externals, stack, func, pop)

    def __exec_unary(self, ctx, externals, stack, func, pop):
        stack.push_intptr(func(pop()))

    def __exec_binary(self, ctx, externals, stack, func, pop):
        b = pop()
        a = pop()
        stack.push_intptr(func(a, b))

    def exec_addr(self, ctx, externals, stack):
        exception = None
        for name in self.operand:
            try:
                value = ctx.env.lookup_symbol(name)
            except KeyError as e:
                exception = e
                continue
            stack.push_intptr(value)
            return
        assert exception is not None
        raise exception

    def __exec_constX(self, ctx, externals, stack):
        stack.push_intptr(self.operand)

    def exec_bra(self, ctx, externals, stack):
        if stack.pop_unsigned() != 0:
            return self.operand

    def exec_deref(self, ctx, externals, stack):
        self.__exec_deref(ctx, externals, stack, 0)

    def exec_deref_int(self, ctx, externals, stack):
        assert self.operand != 0
        self.__exec_deref(ctx, externals, stack, self.operand)

    def __exec_deref(self, ctx, externals, stack, size):
        if size == 0:
            size = ctx.wordsize
        is_signed = size < 0
        if is_signed:
            size *= -1
        size >>= 3
        sizecode = "%s%d" % (is_signed and "s" or "u", size)
        sizecode = self.FIXEDSIZE.get(sizecode, None)
        if sizecode is None:
            raise UnhandledNoteError(self)
        check, fmt = sizecode
        assert check == size
        fmt = self.byteorder + fmt
        try:
            result = ctx.env.read_memory(fmt, stack.pop_unsigned())
        except KeyError as e:
            raise BadDerefError(self, ctx.env.memory, e.args[0])
        stack.push_intptr(struct.unpack(fmt, result)[0])

    def exec_cast_int2ptr(self, ctx, externals, stack):
        pass

    def exec_cast_ptr2int(self, ctx, externals, stack):
        pass

    def exec_drop(self, ctx, externals, stack):
        stack.pop_boxed()

    def exec_dup(self, ctx, externals, stack):
        stack.push_boxed(stack.slots[0])

    def exec_call(self, ctx, externals, stack):
        callee = stack.pop_function()
        ctx.trace_call(callee, stack)
        callee.execute(ctx, stack)

    def __exec_litN(self, ctx, externals, stack):
        stack.push_intptr(self.opcode - constants.DW_OP_lit0)

    def exec_load_external(self, ctx, externals, stack):
        if self.operand == 0:
            external = self.function.type, self.function
        else:
            external = externals[self.operand - 1].resolve(ctx)
        stack.push_typed(*external)

    def exec_over(self, ctx, externals, stack):
        stack.push_boxed(stack.slots[1])

    def exec_pick(self, ctx, externals, stack):
        stack.push_boxed(stack.slots[self.operand])

    def exec_plus_uconst(self, ctx, externals, stack):
        a = stack.pop_unsigned()
        stack.push_intptr(a + self.operand)

    def exec_rot(self, ctx, externals, stack):
        a = stack.pop_boxed()
        b = stack.pop_boxed()
        c = stack.pop_boxed()
        stack.push_boxed(a)
        stack.push_boxed(c)
        stack.push_boxed(b)

    def exec_skip(self, ctx, externals, stack):
        return self.operand

    def exec_swap(self, ctx, externals, stack):
        a = stack.pop_boxed()
        b = stack.pop_boxed()
        stack.push_boxed(a)
        stack.push_boxed(b)

    def exec_warn(self, ctx, externals, stack):
        ctx.env.warn_caller(self.operand)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import ParserError, ParsedError, StackTypeError

SOURCE = """\
typedef ptr ptr_alias

define test::deref
    argument %s arg

    deref %s
"""

class TestDeref(TestCase):
    TYPES = ("ptr", "ptr_alias",
             "int", "bool",
             "opaque", "func ()", "func int (ptr)",
             "uint8_t", "uint16_t", "uint32_t", "uint64_t",
             "int8_t", "int16_t", "int32_t", "int64_t",
             "intptr_t", "uintptr_t")

    def test_deref(self):
        """Check that deref works."""
        for argtype in self.TYPES:
            argtype_is_ok = argtype.startswith("ptr")
            for rettype in self.TYPES:
                rettype_is_func = rettype.startswith("func")
                rettype_is_ok = (rettype.startswith("ptr")
                                 or rettype.endswith("_t"))

                source = SOURCE % (argtype, rettype)

                if rettype_is_func:
                    exception = ParserError
                elif not rettype_is_ok:
                    exception = ParsedError
                elif not argtype_is_ok:
                    exception = StackTypeError
                else:
                    exception = None

                if exception is not None:
                    self.assertRaises(exception, self.compile, source)
                    continue

                tree, output = self.compile(source)
                ops = output.ops
                self.assertEqual(len(ops), 1)
                op = ops[0]

                if rettype in ("ptr", "ptr_alias"):
                    self.assertEqual("deref", op.name)
                    continue

                self.assertEqual("deref_int", op.name)
                is_signed = rettype[0] != "u"
                if not is_signed:
                    rettype = rettype[1:]
                rettype = rettype[3:-2]
                if rettype == "ptr":
                    sizecode = self._wordsize
                else:
                    sizecode = int(rettype)
                self.assertNotEqual(sizecode, 0)
                if is_signed:
                    sizecode *= -1
                self.assertEqual(op.operand, sizecode)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

NT_GNU_INFINITY = 5

I8_CHUNK_SIGNATURE = 1
I8_CHUNK_BYTECODE = 2
I8_CHUNK_EXTERNALS = 3
I8_CHUNK_STRINGS = 4
I8_CHUNK_CODEINFO = 5

I8_TYPE_INT = "i"
I8_TYPE_PTR = "p"
I8_TYPE_OPAQUE = "o"
I8_TYPE_FUNC = "F"

DW_OP_addr = 0x03
DW_OP_deref = 0x06
DW_OP_const1u = 0x08
DW_OP_const1s = 0x09
DW_OP_const2u = 0x0a
DW_OP_const2s = 0x0b
DW_OP_const4u = 0x0c
DW_OP_const4s = 0x0d
DW_OP_const8u = 0x0e
DW_OP_const8s = 0x0f
DW_OP_constu = 0x10
DW_OP_consts = 0x11
DW_OP_dup = 0x12
DW_OP_drop = 0x13
DW_OP_over = 0x14
DW_OP_pick = 0x15
DW_OP_swap = 0x16
DW_OP_rot = 0x17
DW_OP_xderef = 0x18
DW_OP_abs = 0x19
DW_OP_and = 0x1a
DW_OP_div = 0x1b
DW_OP_minus = 0x1c
DW_OP_mod = 0x1d
DW_OP_mul = 0x1e
DW_OP_neg = 0x1f
DW_OP_not = 0x20
DW_OP_or = 0x21
DW_OP_plus = 0x22
DW_OP_plus_uconst = 0x23
DW_OP_shl = 0x24
DW_OP_shr = 0x25
DW_OP_shra = 0x26
DW_OP_xor = 0x27
DW_OP_bra = 0x28
DW_OP_eq = 0x29
DW_OP_ge = 0x2a
DW_OP_gt = 0x2b
DW_OP_le = 0x2c
DW_OP_lt = 0x2d
DW_OP_ne = 0x2e
DW_OP_skip = 0x2f
DW_OP_lit0 = 0x30
DW_OP_lit1 = 0x31
DW_OP_lit2 = 0x32
DW_OP_lit3 = 0x33
DW_OP_lit4 = 0x34
DW_OP_lit5 = 0x35
DW_OP_lit6 = 0x36
DW_OP_lit7 = 0x37
DW_OP_lit8 = 0x38
DW_OP_lit9 = 0x39
DW_OP_lit10 = 0x3a
DW_OP_lit11 = 0x3b
DW_OP_lit12 = 0x3c
DW_OP_lit13 = 0x3d
DW_OP_lit14 = 0x3e
DW_OP_lit15 = 0x3f
DW_OP_lit16 = 0x40
DW_OP_lit17 = 0x41
DW_OP_lit18 = 0x42
DW_OP_lit19 = 0x43
DW_OP_lit20 = 0x44
DW_OP_lit21 = 0x45
DW_OP_lit22 = 0x46
DW_OP_lit23 = 0x47
DW_OP_lit24 = 0x48
DW_OP_lit25 = 0x49
DW_OP_lit26 = 0x4a
DW_OP_lit27 = 0x4b
DW_OP_lit28 = 0x4c
DW_OP_lit29 = 0x4d
DW_OP_lit30 = 0x4e
DW_OP_lit31 = 0x4f
DW_OP_reg0 = 0x50
DW_OP_reg1 = 0x51
DW_OP_reg2 = 0x52
DW_OP_reg3 = 0x53
DW_OP_reg4 = 0x54
DW_OP_reg5 = 0x55
DW_OP_reg6 = 0x56
DW_OP_reg7 = 0x57
DW_OP_reg8 = 0x58
DW_OP_reg9 = 0x59
DW_OP_reg10 = 0x5a
DW_OP_reg11 = 0x5b
DW_OP_reg12 = 0x5c
DW_OP_reg13 = 0x5d
DW_OP_reg14 = 0x5e
DW_OP_reg15 = 0x5f
DW_OP_reg16 = 0x60
DW_OP_reg17 = 0x61
DW_OP_reg18 = 0x62
DW_OP_reg19 = 0x63
DW_OP_reg20 = 0x64
DW_OP_reg21 = 0x65
DW_OP_reg22 = 0x66
DW_OP_reg23 = 0x67
DW_OP_reg24 = 0x68
DW_OP_reg25 = 0x69
DW_OP_reg26 = 0x6a
DW_OP_reg27 = 0x6b
DW_OP_reg28 = 0x6c
DW_OP_reg29 = 0x6d
DW_OP_reg30 = 0x6e
DW_OP_reg31 = 0x6f
DW_OP_breg0 = 0x70
DW_OP_breg1 = 0x71
DW_OP_breg2 = 0x72
DW_OP_breg3 = 0x73
DW_OP_breg4 = 0x74
DW_OP_breg5 = 0x75
DW_OP_breg6 = 0x76
DW_OP_breg7 = 0x77
DW_OP_breg8 = 0x78
DW_OP_breg9 = 0x79
DW_OP_breg10 = 0x7a
DW_OP_breg11 = 0x7b
DW_OP_breg12 = 0x7c
DW_OP_breg13 = 0x7d
DW_OP_breg14 = 0x7e
DW_OP_breg15 = 0x7f
DW_OP_breg16 = 0x80
DW_OP_breg17 = 0x81
DW_OP_breg18 = 0x82
DW_OP_breg19 = 0x83
DW_OP_breg20 = 0x84
DW_OP_breg21 = 0x85
DW_OP_breg22 = 0x86
DW_OP_breg23 = 0x87
DW_OP_breg24 = 0x88
DW_OP_breg25 = 0x89
DW_OP_breg26 = 0x8a
DW_OP_breg27 = 0x8b
DW_OP_breg28 = 0x8c
DW_OP_breg29 = 0x8d
DW_OP_breg30 = 0x8e
DW_OP_breg31 = 0x8f
DW_OP_regx = 0x90
DW_OP_fbreg = 0x91
DW_OP_bregx = 0x92
DW_OP_piece = 0x93
DW_OP_deref_size = 0x94
DW_OP_xderef_size = 0x95
DW_OP_nop = 0x96
DW_OP_push_object_address = 0x97
DW_OP_call2 = 0x98
DW_OP_call4 = 0x99
DW_OP_call_ref = 0x9a
DW_OP_form_tls_address = 0x9b
DW_OP_call_frame_cfa = 0x9c
DW_OP_bit_piece = 0x9d
DW_OP_implicit_value = 0x9e
DW_OP_stack_value = 0x9f

# GNU extensions
DW_OP_GNU_push_tls_address = 0xe0  # same as DW_OP_HP_unknown
DW_OP_GNU_uninit = 0xf0
DW_OP_GNU_encoded_addr = 0xf1
DW_OP_GNU_implicit_pointer = 0xf2
DW_OP_GNU_entry_value = 0xf3
DW_OP_GNU_const_type = 0xf4
DW_OP_GNU_regval_type = 0xf5
DW_OP_GNU_deref_type = 0xf6
DW_OP_GNU_convert = 0xf7
DW_OP_GNU_reinterpret = 0xf9
DW_OP_GNU_parameter_ref = 0xfa
DW_OP_GNU_addr_index = 0xfb
DW_OP_GNU_const_index = 0xfc
DW_OP_GNU_wide_op = 0xff

# HP extensions
DW_OP_HP_unknown = 0xe0  # same as GNU_push_tls_address
DW_OP_HP_is_value = 0xe1
DW_OP_HP_fltconst4 = 0xe2
DW_OP_HP_fltconst8 = 0xe3
DW_OP_HP_mod_range = 0xe4
DW_OP_HP_unmod_range = 0xe5
DW_OP_HP_tls = 0xe6

# PGI (STMicroelectronics) extensions
DW_OP_PGI_omp_thread_num = 0xf8

# Infinity wide operations
I8_OP_call = 0x100
I8_OP_load_external = 0x101
I8_OP_deref_int = 0x102
I8_OP_cast_int2ptr = 0x103
I8_OP_cast_ptr2int = 0x104
I8_OP_warn = 0x105



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import RedefinedIdentError
from . import names
from . import parser
from . import stack
from . import types
from . import visitors
from . import warn
import copy

class External(stack.Element):
    def __init__(self, name, type, ast):
        if type.basetype is types.PTRTYPE:
            assert name.is_shortname
        else:
            assert type.is_function
            assert name.is_fullname
        stack.Element.__init__(self, type, name)
        self.ast = ast

    @property
    def name(self):
        return self.names[0]

class PlaceHolder(object):
    def __init__(self, name):
        self.name = name

class ExternTable(visitors.Visitable):
    def __init__(self):
        self.default_provider = None
        self.entries = {}

    def add(self, node, type):
        """Add an entry to the table."""
        self.__add(node, External, type, node)

    def block(self, node):
        """Reserve a name in the table."""
        self.__add(node, PlaceHolder)

    def __add(self, node, klass, *args):
        entry = klass(node.name.value, *args)
        key = str(entry.name)

        prev = entry.name
        if prev.is_fullname and prev.provider == self.default_provider:
            prev = prev.without_provider()
        prev = self.__lookup(prev)
        if prev is not None:
            if (isinstance(entry, External)
                  and isinstance(prev, External)
                  and entry.type.encoding == prev.type.encoding):
                if not isinstance(entry.ast, parser.Function):
                    warn("unnecessary definition of ‘%s’" % entry.name,
                         entry.name)
                    warn("‘%s’ was first defined here" % prev.name,
                         prev.name)
                return
            raise RedefinedIdentError(entry.name, "name", key, prev.name)

        self.entries[key] = entry

    def lookup(self, name):
        result = self.__lookup(name)
        if isinstance(result, PlaceHolder):
            result = None
        return result

    def __lookup(self, name):
        result = self.__lookup_one(name)
        if result is None and name.is_shortname:
            fullname = name.with_provider(self.default_provider)
            result = self.__lookup_one(fullname)
        return result

    def __lookup_one(self, name):
        assert isinstance(name, names.Name)
        return self.entries.get(str(name), None)

class TableCreator(object):
    def visit_external(self, external):
        self.table.add(external, external.typename.type)

class PerFileTableCreator(TableCreator):
    def visit_toplevel(self, toplevel):
        toplevel.table = self.table = ExternTable()
        for node in toplevel.children:
            node.accept(self)

    def visit_function(self, function):
        self.table.add(function, function.type)

    def visit_wordsize(self, wordsize):
        pass

    def visit_typedef(self, typedef):
        pass

class PerFuncTableCreator(TableCreator):
    def visit_toplevel(self, toplevel):
        for node in toplevel.functions:
            self.table = copy.copy(toplevel.table)
            self.table.entries = copy.copy(self.table.entries)
            node.accept(self)

    def visit_function(self, function):
        for node in function.children:
            node.accept(self)
        function.externals = self.table

    def visit_fullname(self, name):
        self.table.default_provider = name.provider

    def visit_parameters(self, parameters):
        for node in parameters.children:
            node.accept(self)

    def visit_parameter(self, parameter):
        self.table.block(parameter)

    def visit_returntypes(self, returntypes):
        pass

    def visit_externals(self, externals):
        for node in externals.children:
            node.accept(self)

    def visit_operations(self, ops):
        pass




# -*- coding: utf-8 -*-
# Copyright (C) 2016 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import ParserError
import itertools

LINES = ("wordsize %s",
         "typedef int size_t",
         "extern ptr __stack_user",
         "extern func pid_t () procservice::getpid",
         "define test::test_wordsize returns int")

class TestParseWordsize(TestCase):
    """Check that wordsize is only allowed at the start of the input."""

    def test_parse_wordsize(self):
        """Check that wordsize is parsed correctly"""
        for lines in itertools.permutations(LINES):
            srcfmt = "\n".join(lines) + "\n"
            for wordsize in (32, "word", "extern", "typedef", "define"):
                source = srcfmt % wordsize
                # Test with one wordsize directive
                self.__one_test(source, True)
                # Test with two wordsize directives
                self.__one_test("wordsize 32\n" + source, False)

    def __one_test(self, source, maybe_valid):
        is_valid = maybe_valid and source.startswith("wordsize 32\n")
        if not is_valid:
            self.assertRaises(ParserError, self.compile, source)
        else:
            try:
                self.compile(source)
            except ParserError as e:
                self.fail("the parser should accept this")
            except:
                pass
            else:
                self.fail("should have failed somewhere in the compiler")



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import integer
from .. import archspec
from .. import constants
from . import *
from . import leb128
from . import operations
from . import types
import struct

class Function(object):
    def __init__(self, src):
        self.src = src
        self.__sig_set = False

    @property
    def byteorder(self):
        return self.src.byteorder

    def set_signature(self, provider, name, ptypes, rtypes):
        assert not self.__sig_set
        self.provider = provider
        self.name = name
        self.ptypes = ptypes
        self.rtypes = rtypes
        self.type = types.FunctionType(ptypes, rtypes)
        self.__sig_set = True

    @property
    def signature(self):
        return "%s::%s(%s)%s" % (self.provider, self.name,
                                 types.encode(self.ptypes),
                                 types.encode(self.rtypes))

    def __str__(self):
        return self.signature

class BuiltinFunction(Function):
    """A function provided by the note consumer."""

    def __init__(self, ref, impl):
        Function.__init__(self, ref)
        self.set_signature(ref.provider, ref.name, ref.ptypes, ref.rtypes)
        self.impl = impl

    def execute(self, ctx, stack):
        args = stack.pop_multi(reversed(self.ptypes))
        args.reverse()
        result = self.impl(*args)
        if len(self.rtypes) == 0:
            assert result is None
            result = []
        elif len(self.rtypes) == 1:
            result = [result]
        stack.push_multi(self.rtypes, result)

class BytecodeFunction(Function):
    def __init__(self, src):
        Function.__init__(self, src)
        self.__split_chunks()
        self.__unpack_signature()
        self.__unpack_codeinfo()
        self.__unpack_bytecode()
        self.__unpack_externals()

    def __split_chunks(self):
        self.chunks, offset = {}, 0
        while offset < len(self.src):
            start = offset
            offset, type_id = leb128.read_uleb128(self.src, offset)
            offset, version = leb128.read_uleb128(self.src, offset)
            offset, size = leb128.read_uleb128(self.src, offset)
            limit = offset + size
            if type_id not in self.chunks:
                self.chunks[type_id] = []
            chunk = self.src[offset:limit]
            chunk.version = version
            self.chunks[type_id].append(chunk)
            offset = limit
        if offset != len(self.src):
            raise CorruptNoteError(self.src)

    def one_chunk(self, type_id, supported_versions, is_mandatory):
        if isinstance(supported_versions, integer):
            supported_versions = (supported_versions,)
        chunks = self.chunks.get(type_id, None)
        if chunks is None:
            if not is_mandatory:
                return
            raise UnhandledNoteError(self.src)
        if len(chunks) != 1:
            # The second chunk is the first error
            raise UnhandledNoteError(chunks[1])
        chunk = chunks[0]
        if chunk.version not in supported_versions:
            raise UnhandledNoteError(chunk)
        return chunk

    def get_string(self, start):
        chunk = self.one_chunk(constants.I8_CHUNK_STRINGS, 1, True)
        unterminated = chunk + start
        limit = unterminated.bytes.find(b"\0")
        if limit < 0:
            raise CorruptNoteError(unterminated)
        return unterminated[:limit]

    def __unpack_signature(self):
        chunk = self.one_chunk(constants.I8_CHUNK_SIGNATURE, 2, True)

        offset, provider_o = leb128.read_uleb128(chunk, 0)
        offset, name_o = leb128.read_uleb128(chunk, offset)
        offset, ptypes_o = leb128.read_uleb128(chunk, offset)
        offset, rtypes_o = leb128.read_uleb128(chunk, offset)

        provider, name, ptypes, rtypes \
            = map(self.get_string,
                  (provider_o, name_o, ptypes_o, rtypes_o))

        ptypes = types.decode(ptypes)
        rtypes = types.decode(rtypes)
        self.set_signature(provider.text, name.text, ptypes, rtypes)

    def __unpack_codeinfo(self):
        chunk = self.one_chunk(constants.I8_CHUNK_CODEINFO, 1, False)
        if chunk is None:
            return

        expect = archspec.encode(chunk.wordsize)
        format = self.byteorder + b"H"
        offset = struct.calcsize(format)
        actual = struct.unpack(format, chunk[:offset].bytes)[0]
        if actual != expect:
            raise UnhandledNoteError(chunk)

        offset, self.max_stack = leb128.read_uleb128(chunk, offset)

    def __unpack_bytecode(self):
        self.ops = {}

        chunk = self.one_chunk(constants.I8_CHUNK_BYTECODE, 3, False)
        if chunk is None:
            return

        self.bytecode = chunk
        pc, limit = 0, len(self.bytecode)
        while pc < limit:
            op = operations.Operation(self, pc)
            self.ops[pc] = op
            pc += op.size
        if pc != limit:
            raise CorruptNoteError(self.bytecode + pc)

    def __unpack_externals(self):
        self.externals = []

        chunk = self.one_chunk(constants.I8_CHUNK_EXTERNALS, 2, False)
        if chunk is None:
            return

        unterminated = chunk
        while len(unterminated):
            extern = UnresolvedFunction(self, unterminated)
            self.externals.append(extern)
            unterminated += len(extern.src)

    @property
    def external_functions(self):
        return [str(ext)
                for ext in self.externals
                if isinstance(ext, UnresolvedFunction)]

    def execute(self, ctx, caller_stack):
        stack = ctx.new_stack()
        caller_stack.pop_multi_onto(reversed(self.ptypes), stack)
        pc, return_pc = 0, len(self.bytecode)
        while pc >= 0 and pc < return_pc:
            op = self.ops.get(pc, None)
            if op is None:
                raise BadJumpError(stack.op)
            stack.op = op
            pc_adjust = op.execute(ctx, self.externals, stack)
            pc += op.size
            if pc_adjust is not None:
                pc += pc_adjust
            last_op = op
        if pc != return_pc:
            raise BadJumpError(stack.op)
        ctx.trace_return((self.signature, pc), stack)
        stack.pop_multi_onto(self.rtypes, caller_stack)

    @property
    def coverage(self):
        ops_hit = opcount = 0
        for op in self.ops.values():
            if op.hitcount > 0:
                ops_hit += 1
            opcount += 1
        return ops_hit, opcount

class UnresolvedFunction(Function):
    def __init__(self, referrer, unterminated):
        offset, provider_o = leb128.read_uleb128(unterminated, 0)
        offset, name_o = leb128.read_uleb128(unterminated, offset)
        offset, ptypes_o = leb128.read_uleb128(unterminated, offset)
        offset, rtypes_o = leb128.read_uleb128(unterminated, offset)
        Function.__init__(self, unterminated[:offset])

        provider, name, ptypes, rtypes \
            = map(referrer.get_string,
                  (provider_o, name_o, ptypes_o, rtypes_o))

        ptypes = types.decode(ptypes)
        rtypes = types.decode(rtypes)
        self.set_signature(provider.text, name.text, ptypes, rtypes)

    def resolve(self, ctx):
        return self.type, ctx.get_function(self)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

SOURCE = """\
define test::optimize_use_plus_uconst returns int
    argument int x

    load %s
    add
"""

class TestOptimizeUsePlusUconst(TestCase):
    def test_optimize_use_plus_uconst(self):
        """Check that DW_OP_plus_uconst is used where possible."""
        for value in ("TRUE", 5, 255, 16384, 327889, 123456789):
            tree, output = self.compile(SOURCE % value)
            ops = output.ops
            self.assertEqual(len(ops), 1)
            self.assertEqual(ops[0].name, "plus_uconst")
            self.assertEqual(ops[0].operand, value == "TRUE" and 1 or value)

    def test_optimize_dont_use_plus_uconst_neg(self):
        """Check that DW_OP_plus_uconst is not used for negative numbers."""
        for value in (-1, -5, -255, -16384, -327889, -123456789):
            tree, output = self.compile(SOURCE % value)
            ops = output.ops
            self.assertEqual(len(ops), 2)
            self.assertEqual(ops[0].operand, value)
            self.assertEqual(ops[1].name, "plus")

    def test_optimize_dont_use_plus_uconst_zero(self):
        """Check that DW_OP_plus_uconst is not used for zero."""
        for value in (0, "FALSE"):
            tree, output = self.compile(SOURCE % value)
            self.assertEqual(len(output.ops), 0)
        tree, output = self.compile(SOURCE.replace("int", "ptr") % value)
        self.assertEqual(len(output.ops), 0)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import InvalidCastError
from i8c.compiler import ParserError
from i8c.compiler import StackError
from i8c.compiler import UndefinedIdentError
from i8c.compiler import UnnecessaryCastError

SOURCE = """\
typedef func ptr (int) derived_func_t
typedef ptr derived_ptr_t
typedef int derived_int_t
typedef opaque derived_opaque_t

define test::test_cast returns %s
   argument %s an_argument
"""

TYPEGROUPS = (
    ("int",             # builtin
     "bool",            # builtin (derived)
     "int32_t",         # builtin (derived, sized)
     "derived_int_t"),  # user-defined (derived)

    ("ptr",             # builtin
     "derived_ptr_t"),  # user-defined (derived)

    ("opaque",              # builtin
     "derived_opaque_t"),   # user-defined (derived)

    ("func ptr (int)",  # builtin (function)
     "derived_func_t")) # user-defined (function, derived)

class TestCast(TestCase):
    TYPES = []
    CASTS = {}
    for group in TYPEGROUPS:
        for type1 in group:
            TYPES.append(type1)
            for type2 in group:
                CASTS[(type1, type2)] = True
    del group, type1, type2

    @classmethod
    def types_match(cls, type1, type2):
        return cls.CASTS.get((type1, type2), False)

    @classmethod
    def can_cast(cls, type1, type2):
        for ab in (("int", "ptr"), ("ptr", "int")):
            if (cls.types_match(type1, ab[0])
                  and cls.types_match(type2, ab[1])):
                return True
        return False

    def test_cast(self):
        """Check that cast works."""
        for rtype in self.TYPES:
            for ptype in self.TYPES:
                # Test without casting
                source = SOURCE % (rtype, ptype)
                if self.types_match(ptype, rtype):
                    error = None
                else:
                    error = StackError
                self.__do_test(source, error)

                # Test with a cast
                for slot in (0, "an_argument", 15, "no_such_slot"):
                    cast = "cast %s, %s" % (slot, rtype)
                    if rtype == "func ptr (int)":
                        error = ParserError
                    elif slot == 15:
                        error = StackError
                    elif slot == "no_such_slot":
                        error = UndefinedIdentError
                    elif self.types_match(ptype, rtype):
                        error = UnnecessaryCastError
                    elif self.can_cast(ptype, rtype):
                        error = None
                    else:
                        error = InvalidCastError
                    self.__do_test(source + cast, error)

    def __do_test(self, source, error):
        if error is None:
            tree, output = self.compile(source)
        else:
            self.assertRaises(error, self.compile, source)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import fprint, integer
from . import functions
from . import memory
from . import types
import ctypes
import sys

# XXX most every assert here should be a proper error

class Stack(object):
    def __init__(self, wordsize):
        self.__sint_t = getattr(ctypes, "c_int%d" % wordsize)
        self.__uint_t = getattr(ctypes, "c_uint%d" % wordsize)
        self.slots = []

    # Generic push and pop

    def pop_multi_onto(self, types, other):
        """Transfer values from one stack to another."""
        types = list(types)
        other.push_multi(types, self.pop_multi(types))

    def push_multi(self, types, values):
        types, values = map(list, (types, values))
        assert len(types) == len(values)
        typedvalues = list(zip(types, values))
        typedvalues.reverse()
        for type, value in typedvalues:
            self.push_typed(type, value)

    def pop_multi(self, types):
        result = []
        for type in types:
            result.append(self.pop_typed(type))
        return result

    def push_typed(self, type, value):
        assert type is not None
        self.push_boxed(self.__box(type, value))

    def pop_typed(self, type):
        assert type is not None
        return self.__unbox(type, self.pop_boxed())

    def push_boxed(self, boxed_value):
        assert not isinstance(boxed_value, integer)
        self.slots.insert(0, boxed_value)

    def pop_boxed(self):
        return self.slots.pop(0)

    # Push and pop shortcuts

    def push_intptr(self, value):
        self.push_typed(types.IntPtrType, value)

    def pop_unsigned(self):
        return self.pop_typed(types.IntPtrType)

    def pop_signed(self):
        return self.__sint_t(self.pop_unsigned()).value

    def pop_function(self):
        return self.__unbox_FUNCTION(None, self.pop_boxed())

    # Boxing and unboxing

    def __box(self, type, value):
        return getattr(
            self, "_Stack__box_" + type.name.upper())(type, value)

    def __unbox(self, type, value):
        return getattr(
            self, "_Stack__unbox_" + type.name.upper())(type, value)

    def __box_INTPTR(self, type, value):
        if isinstance(value, memory.Block):
            value = value.location
        assert isinstance(value, integer)
        return self.__uint_t(value)

    def __unbox_INTPTR(self, type, boxed):
        assert isinstance(boxed, self.__uint_t)
        return boxed.value

    def __box_FUNCTION(self, type, value):
        if not isinstance(value, functions.Function):
            assert callable(value)
            # Wrap anonymous functions provided by the testcase in
            # argument lists.  This is persistent (unboxing won't
            # remove it) so what you pop in this case is not exactly
            # what you pushed.
            value = functions.BuiltinFunction(AnonFuncRef(type), value)
        assert value.type == type
        return value

    def __unbox_FUNCTION(self, type, boxed):
        assert isinstance(boxed, functions.Function)
        assert type is None or boxed.type == type
        return boxed

    def __box_OPAQUE(self, type, value):
        return Opaque(value)

    def __unbox_OPAQUE(self, type, boxed):
        assert isinstance(boxed, Opaque)
        return boxed.value

    # Tracing

    def trace(self, tracelevel):
        depth = tracelevel + 1
        for item, index in zip(self.slots[:depth], range(depth)):
            if isinstance(item, self.__uint_t):
                value = item.value
                item = "%d" % value
                if value < 0 or value > 15:
                    item += " (0x%x)" % value
            fprint(sys.stdout, "    stack[%d] = %s" % (index, item))
        fprint(sys.stdout)

class AnonFuncRef(object):
    def __init__(self, type):
        self.provider = "i8x"
        self.name = "anonymous_function_%x" % id(self)
        self.ptypes = type.ptypes
        self.rtypes = type.rtypes

class Opaque(object):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "OPAQUE [0x%x]" % id(self)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

SOURCE1 = """\
define test::externals_test returns int
    extern ptr a_symbol
    extern func int, int (ptr) a_function
    load a_symbol
    call a_function
"""

SOURCE2 = """\
typedef ptr ptr_alias_t
typedef func int, int (ptr) fun_alias_f

define test::externals_test returns int
    extern ptr_alias_t a_symbol
    extern fun_alias_f a_function
    load a_symbol
    call a_function
"""

class TestExternals(TestCase):
    def test_externals(self):
        """Check that externals work."""
        for source in SOURCE1, SOURCE2:
            tree, output = self.compile(source)
            self.assertEqual(
                ["addr", "load_external", "call"],
                output.opnames)
